//
// gt_main.cpp
//
// Main routines
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with pkthisto in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "stdafx.h"
#define FIRSTTIME
#include "gt.h"


int main(int argc, char* argv[])
{
	char	configfile[128]; // Name of the file from which to read configs
	char	outf_name[128];	// Actual name of base output file

	struct global_state gs;

	memset((char *)(&gs), 0, sizeof(struct global_state));

	// Initialize default filenames

	strcpy(configfile, CONFDEFAULT);
	strcpy(gs.outf_prefix, OUTFDEFAULT);	// output file name prefix
	strcpy(gs.version, PH_VERSION);

	// Initialize default non-zero parameters

	gs.stats |= FILE_STATS;
	gs.flow_max_gap = FLOW_MAX_MS_DEFAULT*1000;
	gs.pkt_cnt_min = PKT_CNT_MIN_DEFAULT;
	gs.pkt_per_histo = PKT_PER_HISTO_DEFAULT;
	gs.interval_histo_width = INTERHISTOWIDTH;
	gs.length_histo_width = LENHISTOWIDTH;
	gs.chkpoint_interval = CHKPOINTINTERVAL;
	gs.last_time_sec = -1;
		
	// Parse command line

	while (argc>1)
	{
	 if (argv[1][0]=='-')
      {
        switch (argv[1][1])
		{
			
		 case 'd':
			{	// Set debug level
				gs.debug = strtol(argv[2],(char **)NULL,0);
			}
			break;

		 case 'c':
			{	// Set alternate name for the config file
				strcpy(configfile, argv[2]);
			}
			break;

		 case 'v':
			{	// Print the version number then quit.
				printf("pkthisto, version %s\n", gs.version);
				exit(-1);
			}
			break;

		 case 'r':
			{
				// Hack function for reconstructing pure ASCII
				// versions of compressed ASCII histogram files.
				// If this option is used, pkthisto exist immediately
				// after performing the conversion.			
				gts_convertfile(argv[2]);
				exit(0);
			}
			break;

		 case 'o':
			{	// Specify the output file prefix (default is 'gt_out')
				strcpy(gs.outf_prefix, argv[2]);
			}
			break;

		 case 'h':
			printf("pkthisto -d <debuglevel> -c <file> -o <file> \n");
			printf("version %s\n", gs.version);
			printf("\t-c <file>  Config file is <file>\n");
			printf("\t-d <level> Debug level is <level>\n");
			printf("\t-o <file>  Output file is <file>\n");
			printf("\t-r <file>  Dump pure ASCII version of compressed ASCII <file>\n");
			printf("\t-v         Print version number\n");
			printf("\n");
#ifdef WIN32
			printf("\tCreate histograms of UDP/IP traffic from tcpdump\n");
			printf("\tor NAI Sniffer tracefiles.\n");
#else
			printf("\tCreate histograms of UDP/IP traffic from real time capture,\n");
			printf("\ttcpdump tracefiles, or NAI Sniffer tracefiles.\n");
#endif
			printf("\t(c)2001 Grenville Armitage, gj_armitage@yahoo.com\n");
			exit(-1);
			break;

		 default:
			 break;
		}

          argv +=2;
          argc -=2;
	 }
	}

	// Open the global output file, which will be referenced
	// throughout the rest of the program as gs->outf. This
	// file will be the repository for general status messages
	// while we're processing packet records.

	sprintf(outf_name, "%s0.txt\0", gs.outf_prefix);
	if ((gs.outf = fopen(outf_name, "w")) == NULL)
	{
		printf("Opening output file %s failed, aborting.\n",
			outf_name);
		exit(1);
	}

	// Read configuration controls from config file

	parse_config_file(&gs, configfile);

	if (sanity_check_config(&gs))
	{
		printf("Failed sanity check, exiting.\n");
		exit(-1);		
	}

	fprintf(gs.outf, "pkthisto version %s\n", gs.version);
	
	// Now either initiate realtime packet snooping from a
	// local interface, or begin reading a tracefile. Return
	// after reading entire file, or after reading gs.total_pkts
	// packets.


	if (FLAG_SET(gs.stats, PKTSNOOP))
#ifndef WIN32
		do_packetsnoop(&gs);
#endif
#ifdef WIN32
		{
			printf("Sorry, the Win32 version does not yet support packet capture\n");
			return(-1);
		}
#endif
	else
		read_tracefiles(&gs);

	
	// We've finished gathering per-flow data structures
	// in memory, performing interim checkpoint dumps along
	// the way, etc. Do a final checkpoint dump.

	gts_checkpoint_stats(&gs, CHKP_FINAL);

	// Along the way, we made estimates of which (if any)
	// IP addresses appeared to represent server nodes
	// (either multiple flows terminated or originated on
	// that address/port combination)
	// Print out that list of nodes we guessed were servers.

	// print_servers(&gs);

	// For now, that's it.

	return(0);

}

//
// read_tracefiles()
//
// Does little but call the tcpdump or sniffer parsing
// routines, based on flags set in config file.
//

void read_tracefiles(struct global_state *gs)
{

	fprintf(gs->outf, "Reading from trace file %s\n", gs->tracefile);
	printf("Reading from trace file %s\n", gs->tracefile);

	if ((gs->in_fd = _open(gs->tracefile, _O_RDONLY | _O_BINARY)) < 0)
		{
		 fprintf(gs->outf, "Opening %s failed: %s\n", gs->tracefile, strerror(errno));
		 printf("Opening %s failed: %s\n", gs->tracefile, strerror(errno));
		 exit(1);
		}

	// Print some basic operating information to the output file
	print_operating_config(gs);

	// Call parser for the desired file format

	if (FLAG_SET(gs->stats, TCPDUMP))
		read_tcpdump_file(gs);

	else if (FLAG_SET(gs->stats, SNIFFER))
		read_naisniffer_file(gs);


	else
	{
		printf("Error: Must set file type. Exiting.\n");
		exit(-1);
	}

}


//
// print_operating_config()
//
// Dump pertinent operating conditions to the primary
// output file
//
void print_operating_config(struct global_state *gs)
{
	struct seen_servers *tmpserv;

	// Print some basic operating information to the output file

	fprintf(gs->outf, "Continuous flow: packets <= %dms apart\n",
		gs->flow_max_gap/1000);
	fprintf(gs->outf, "Histograms generated with at least %d packets\n",
		gs->pkt_cnt_min);
	fprintf(gs->outf, "New histograms every %d packets\n",
		gs->pkt_per_histo);
	fprintf(gs->outf, "Inter-arrival histo: 0 to %d msec, %d usec buckets\n",
		(MAX_INTERARRIVAL*gs->interval_histo_width)/1000,
		gs->interval_histo_width);
	fprintf(gs->outf, "IP pkt len histo: 0 to %d bytes, %d byte buckets\n",
		MAX_LENGTH*gs->length_histo_width,
		gs->length_histo_width);
	fprintf(gs->outf, "Checkpoints every %d minutes\n",
		gs->chkpoint_interval/60);
 
	tmpserv = gs->ssrv;
	while (tmpserv != NULL)
	{
		char serveraddr[32];

		sprintf(serveraddr, "%s", inet_ntoa(tmpserv->srvaddr));
		fprintf(gs->outf, "Server specified as: %s/%d\n",
			serveraddr, tmpserv->srvport);
		printf("Server specified as: %s/%d\n",
			serveraddr, tmpserv->srvport);

		tmpserv = tmpserv->next;
	}

	if (FLAG_SET(gs->stats, COMPRESS_ASCII))
	{
		fprintf(gs->outf, "Compressed-ASCII selected\n");
	}

	if (FLAG_SET(gs->stats, TCPDUMP))
	{
		fprintf(gs->outf, "Reading from TCPDUMP file:\n");
	}

	if (FLAG_SET(gs->stats, SNIFFER))
	{
		fprintf(gs->outf, "Reading from NAI SNIFFER file:\n");
	}

	if (FLAG_SET(gs->stats, PKTSNOOP))
	{
		fprintf(gs->outf, "Beginning real-time traffic capture:\n");
	}

	fflush(gs->outf);

}



//
// gts_convert_oneline()
//
// Takes a single string containing one line
// of compressed ASCII histogram, the number of
// elements, and pointer to an array of
// elements to be filled. Returns with
// each array element containing an integer
// value between 0 and 4095.
//
// Returns an integer value representing the
// total of all bucket contents.
//

int gts_convert_oneline(char *string,
					int num,
					int *array)
{

    int tmp;
    int bucket;
    int total;

	// Zero the elements in the array that we'll
	// be using (assume caller knows how to set hi
	// so it isn't outside bounds of array)
	
	memset((char *)array, 0, num*sizeof(int));

    string++;        // Skip the initial non-space marker
	tmp = 0;
	bucket = 0;
	total = 0;
    // Scan the compressed string until null terminator
    while (*string != 0)
    {
        if (*string == 'z')
        {
                // Extract the digits up to the next "%" to
                // represent how many times to repeat the previous
                // bucket's count value.
                int multiplier, chars = 0;

                sscanf(string,"z%d%%%n", &multiplier, &chars);
                string += chars;
                while (multiplier > 0)
                {
                     array[bucket] += tmp;
                     total += tmp;
                     bucket++;
                     multiplier--;
                }
        }
        else
        {
                // Extract a single 2-digit base64 ASCII value representing
                // the number of elements at the current bucket value
                tmp = (((*string) - 33)*64 + (*(string+1) - 33));
                array[bucket] += tmp;
                total += tmp;
                string += 2;
                bucket++;
        }
    }

    return(total);
}



//
// gts_convertfile()
//
// Takes a filename and treats it like one of
// LH-, CLH-, IH-, or CIH- files in compressed
// ASCII formats, and regenerates a pure ASCII
// equivalent to stdout.
//
// (Note, this routine is called without any of the
// normal pkthisto state established, so the only things
// it can assume are the filename is to be opened, and
// that stdout can be written to.)
//

void gts_convertfile(char *file)
{
	FILE *inf;
	int n;
	int total;
	char type[8];
	char oneline[4096];
	int histonum, low, hi, err;
	int width;
	float scale;
	int *array;
	int arraysize;


	if ((inf = fopen(file, "r")) == NULL)
	{
		printf("Opening %s for input failed, aborting.\n",
			file);
		exit(1);
	}

	arraysize = 0;
	array = NULL;
	
	// Process the file by reading line by line. File
	// type can be determined by the first token, which
	// will be "LH", "IH", "CLH" or "CIH". LH and IH have
	// similar processing, as do CLH and CIH.
	//
 	while (fscanf(inf, "%s ", type) != EOF)
 	{
		if ((strcmp(type, "LH") == 0) ||
				(strcmp(type, "IH") == 0))
		{
			fscanf(inf, "%d %d %d %d %d %[^\r\n]",
					&histonum, &width, &low, &hi,
					&err, oneline);
					
			if (strcmp(type, "LH") == 0)
				scale = (float)width;
			else
				scale = width/(float)1000;

			if (arraysize <= hi/scale)
			{
				// (re)allocate the temporary array space
				if (array != NULL)
					free(array);
					
				if ((array = (int *)malloc(1+sizeof(int)*(int)(hi/scale)))
					== NULL)
					{
						printf("pkthisto: malloc failed, exiting.\n");
						exit(-1);
					}
				arraysize = (int)(hi/scale);
			}
							
			total = err + gts_convert_oneline(oneline, 1+(int)((hi-low)/scale), array);
			
			printf("\n\"%s %d\n", type, histonum);
			
			n = 0;
			while (n <= (hi-low))
			{
				if (strcmp(type, "LH") == 0)
					printf("%d %5.2f\n", (int)((low + n)*scale), 100*(array[n]/(float)total));
				else
					printf("%4.2f %5.2f\n", (float)((low + n)*scale), 100*(array[n]/(float)total));
				
				n++;
			}			
		
		}
    	else if ((strcmp(type, "CLH")== 0) ||
    			(strcmp(type, "CIH") == 0))
    	{
			fscanf(inf, "%d %d %d %d %[^\r\n]",
					&histonum, &width, &low, &hi,
					oneline);
					
			if (strcmp(type, "CLH") == 0)
				scale = (float)width;
			else
				scale = width/(float)1000;

			if (arraysize <= hi/scale)
			{
				// (re)allocate the temporary array space
				if (array != NULL)
					free(array);
					
				if ((array = (int *)malloc(1+sizeof(int)*(int)(hi/scale)))
					== NULL)
					{
						printf("pkthisto: malloc failed, exiting.\n");
						exit(-1);
					}
				arraysize = (int)(hi/scale);
			}
							
			total = err + gts_convert_oneline(oneline, 1+(int)((hi-low)/scale), array);
			
			printf("\n\"%s %d\n", type, histonum);
			
			n = 0;
			while (n <= (hi-low))
			{
				if (strcmp(type, "CLH") == 0)
					printf("%d %5.2f\n", (int)((low + n)*scale), 100*(array[n]/(float)4096));
				else
					printf("%4.2f %5.2f\n", (float)((low + n)*scale), 100*(array[n]/(float)4096));
				
				n++;
			}			
    	}    	
    	else
    		{
    			printf("Error: %s doesn't appear to contain valid pkthisto content\n",
    				file);
    			printf("Line type is: %s\n", type);
    			exit(-1);
    		}
    			
    	}

    // Be tidy - close the file, free allocated temporary array space
    if (inf != NULL)
    	fclose(inf);
    	
    if (array != NULL)
    	free(array);
    	
    return;
}





