//
// gt_nai_sniffer.cpp
//
// Routines to handle specific tracefiles generated by
// NAI's Sniffer program
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with pkthisto in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "stdafx.h"
#include "gt.h"


//
// read_naisniffer_file()
//
// Begin reading the file. For each frame record, call the
// statistics routines to update per-flow structures.
//
// For now, only handles ".enc" format tracefiles
//

void read_naisniffer_file(struct global_state *gs)
{
	struct f_rec_struct frec;
	char initstring[32];
	int	cnt;
	int	fileeof;
	long basetime;	// Start of tracefile, in seconds since 1/1/1970
	int timescale;		// Sniffer files can have variable
						// timescales on the timestamps (not
						// that I allow anything but 2usec scale)


	// Sniffer ".enc" files begin with an ASCII string terminated
	// by a 0x1A. Read this string in and log it.

	cnt = 0;
	memset(initstring, 0, 32);
	while(cnt < 32)
	{
	 if (_read(gs->in_fd, &(initstring[cnt]), 1) != 1)
	 {
		fprintf(gs->outf, "Error reading sniffer file.\n");
		printf("Error reading sniffer file.\n");
		return;
	 }
	 if (initstring[cnt] == 0x1A)
	 {
		 initstring[cnt] = 0;
		 cnt = 32;
	 }
	 else
		 cnt++;
	}

	fprintf(gs->outf, "\nInitial sniffer tracefile string: %s\n",
		initstring);

	// Begin reading records from the tracefile. Should
	// start with a version record - print out information
	// contained therein. Should continue with repeating sequence
	// of packet frame records. Pass the essential information
	// to the statistics routines.

	fileeof = 1;
	while (fileeof != 0)
	{

		// Read in the record's type & length
		if (_read(gs->in_fd, &frec, sizeof(struct f_rec_struct)) < sizeof(struct f_rec_struct))
		{
			fprintf(gs->outf, "Error reading f_rec_struct: %s\n", strerror(errno));
			exit(-1);
		}

		switch (frec.type)
		{
			case REC_VERS:
					// Extract various stats, including tracefile basetime
					parse_nai_version_record(gs, &timescale, &basetime);
				break;

			case REC_FRAME2:
				{
					// Extract the actual frame information, pass it
					// to the statistics routines.

					if (parse_nai_frame_record(gs, &timescale, basetime) == -1)
					{
						perror("ERROR reading frame record - Aborting file read ");
						exit(-1);
					}
				}
				break;
			case REC_EOF:
				fileeof = 0;	// exit while() loop
				break;
			default:
				printf("Unknown record type %d - skipping %d bytes\n",
						frec.type, frec.length);
				lseek(gs->in_fd, frec.length, SEEK_CUR);
				break;
		}
	}
}





// parse_nai_version_record()
//
// Extract information from the trace file corresponding
// to a version record.
//
int parse_nai_version_record(struct global_state *gs,
							 int *tscale,
							 long *baset)
{
	struct f_vers_struct fvers;
	unsigned int trcdate, trctime;
	struct tm tmnow;

	*tscale = 1;

	// Read the version structure

	memset((char *)&fvers, 0, sizeof(struct f_vers_struct));

	if (_read(gs->in_fd, &fvers, sizeof(struct f_vers_struct))
		< sizeof(struct f_vers_struct))
		return (-1);

	fprintf(gs->outf, "Major version %d, Minor version %d\n",
		fvers.maj_vers, fvers.min_vers);

	if ((fvers.cmprs_level == 0) && (fvers.cmprs_ver == 0))
		fprintf(gs->outf, "No compression\n");
	else
		fprintf(gs->outf, "Compression level/version: %d/%d\n",
			fvers.cmprs_level, fvers.cmprs_ver);

	fprintf(gs->outf, "Network type: %d   Following records type: %d\n",
			fvers.network, fvers.type);

	fprintf(gs->outf, "Timeunits: %d\n", fvers.timeunit);

	if (fvers.timeunit != TIMEUNIT_2USEC)
		fprintf(gs->outf, "\nTIMEUNITS NOT RECOGNIZED - Results will be erroneous!\n\n");

	*tscale = 2;	// Two microseconds per timestamp unit is what we're
					// assuming here.

	trcdate = (fvers.date[3] << 8) | fvers.date[2];
	trctime = (fvers.date[1] << 8) | fvers.date[0];

	memset((char *)&tmnow, 0, sizeof(struct tm));
	tmnow.tm_isdst = -1; // Force system to decide if daylight savings was on
	tmnow.tm_sec = trctime & 0x1F;
	tmnow.tm_hour = ((trctime >> 11) & 0x1F);
	tmnow.tm_min = (trctime >> 5) & 0x3F;
	tmnow.tm_year = (trcdate >> 9) + 80; // trcdate is years relative to 1980, whereas
										// tm_year is relative to 1900.
	tmnow.tm_mon = ((trcdate >> 5)& 0xF) - 1;	// NB. In tm_mon january is month "0" not "1"
	tmnow.tm_mday = (trcdate & 0x1F);

	(void)mktime(&tmnow);	// MUST call mktime() before asctime() to ensure
							// fields in tmnow are adjusted correctly

	printf("Trace begins at %s", asctime(&tmnow));
	fprintf(gs->outf, "Trace begins at %s", asctime(&tmnow));

	// Construct a basetime in terms of seconds from 1/1/1970.
	// Each frame record in a sniffer ".enc" file contains timestamps
	// relative to midnight of the day on which the trace begins.

	memset((char *)&tmnow, 0, sizeof(struct tm));
	tmnow.tm_isdst = -1; // Force system to decide if daylight savings was on
	tmnow.tm_sec = 0;
	tmnow.tm_hour = 0;
	tmnow.tm_min = 0;
	tmnow.tm_year = (trcdate >> 9) + 80; // trcdate is years relative to 1980, whereas
										// tm_year is relative to 1900.

	tmnow.tm_mon = ((trcdate >> 5)& 0xF) - 1;	// NB. In tm_mon january is month "0" not "1"
	tmnow.tm_mday = (trcdate & 0x1F);

	*baset = mktime(&tmnow);

	fflush(gs->outf);
	return(0);

}



//
// parse_nai_frame_record()
//
// Extract information from the trace file corresponding
// to a frame record. Extract relevant IP information and
// pass a packet record to the statistics routines.
//

int parse_nai_frame_record(struct global_state *gs,
							 int *tscale,
							 long baset)
{
	struct f_frame2_struct frame;
	struct pkt_record prec;
	long t_sec;
	__int64 timestamp;
	unsigned char pkt[PKT_SIZE_MIN];

	// Read in the f_frame2_struct structure

	memset((char *)&frame, 0, sizeof(struct f_frame2_struct));

	if (_read(gs->in_fd, &frame, sizeof(struct f_frame2_struct))
		< sizeof(struct f_frame2_struct))
	{
		fprintf(gs->outf, "f_frame2_struct read: less than %d bytes. Aborting.\n",
			sizeof(struct f_frame2_struct));
		return (-1);
	}

	if (frame.size >= PKT_SIZE_MIN)
	{
		// The captured length of Ethernet+IP must be
		// at least 38 bytes in order to cover the IP addr
		// and UDP port fields. Proceed to read in the next
		// 38 octets and extract the IP information to pass
		// to the statistics routines.

		if (_read(gs->in_fd, &pkt, PKT_SIZE_MIN) < PKT_SIZE_MIN)
		{
			fprintf(gs->outf,
				"parse_nai_frame_record(): EOF or Error reading packet_header: %s\n",
				strerror(errno));
			printf("parse_nai_frame_record(): EOF or Error reading packet_header: %s\n",
				strerror(errno));
			return(-1);
		}

		// Fill the pkt_record

		// IP src addr is three 'words', or 12 bytes, from beginning
		// of IP header - i.e. 26 bytes from start of pkt[]

		memcpy(&(prec.srcaddr.s_addr), &(pkt[26]), 4);
		memcpy(&(prec.dstaddr.s_addr), &(pkt[30]), 4);
		prec.srcport = (pkt[34] & 0xFF)<<8  |  pkt[35];
		prec.dstport = (pkt[36] & 0xFF)<<8  |  pkt[37];

		// IP packet is 14 bytes shorter than Ethernet frame
		if (frame.true_size != 0)
			prec.len = frame.true_size - 14;
		else
			prec.len = frame.size - 14;

		// Reconstruct the timestamp from the six octets in the
		// frame record header (a 40 bit timestamp value)

		timestamp = (__int64)(frame.time_high & 0xFF);
		timestamp = timestamp << 32;	// Shift up into bits 32 to 39
		timestamp += (__int64)((__int64)frame.time_mid << 16);
		timestamp += (__int64)frame.time_low;
		timestamp = timestamp*(*tscale);		// Adjust to raw usec

		// Adjusting the timestamp to sec/usec since 1/1/1970

		t_sec = (long)(timestamp/1000000);
		prec.time_usec = (long)(timestamp - (1000000*(__int64)t_sec));
		prec.time_sec = baset + t_sec + frame.time_day*86400;

		// Call the statistics routine with this 'event'

		gts_add_pkt_record(gs, &prec);

		lseek(gs->in_fd, frame.size - PKT_SIZE_MIN, SEEK_CUR);
	}
	else
		lseek(gs->in_fd, frame.size, SEEK_CUR);

	return(0);
}
