//
// gt_pktsnoop.cpp
//
// Routines to establish a real-time packet capture
// feed from a local host interface into the statistics
// routines.
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with pkthisto in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "stdafx.h"
#include "gt.h"


#ifndef WIN32	// As yet I haven't written a Win32 version

// Bit of a hack - define a global version of pointer to
// the current state structure so that the signal handlers
// can set the checkpoint flag when called. This pointer is
// set in init_sighandler() and used in usr1_sighandler()
// and term_sighandler().

struct global_state *gs_global;

//
// Initially this is designed to use the BPF driver,
// so it is somewhat system-specific. Basically I'll
// just open the BPF driver and sit in a loop reading
// packets until gs->total_pkts is reached.
//

//
// do_packetsnoop()
//
// Top level entry point is here. Initialize the
// bpf filter/interface then go into a loop reading
// packets. Quit when gs->total_pkts packets are
// read, or SIGTERM is trapped (graceful quit, with
// final checkpoint). If SIGUSR1 is received, do an
// interim checkpoint then resume scanning for packets.
// Otherwise, perform an interim checkpoint every
// gs->chkpoint_interval seconds.

void do_packetsnoop(struct global_state *gs)
{
	int bpfd;
	char *rawfrm;			// Data returned by read() from bpf device
	char *pkt;				// Pointer to ethernet pkt in the frame
	struct bpf_hdr *frame;
	int capend;
	struct pkt_record prec;
	int pktcnt;

	if (gs->snooplink[0] == 0)
	{
		printf("Cannot proceed to packet capture without valid\n");
		printf("link interface name.\n");
		return;
	}
	
	init_sighandler(gs);
	
	// Print some basic operating information to the output file
	print_operating_config(gs);

	// Allocate word-aligned space for reading raw frames
	
	if ((rawfrm = (char *)malloc(BPF_READBUF)) == NULL)
	{
		printf("do_packetsnoop: malloc failed, exiting.\n");
		exit(-1);
	}
	
	// Open the BPF device, attach it to link
	// interface specified by user in config file

	bpfd = init_bpf(gs);
	
	if (bpfd < 0)
	{
		printf("do_packetsnoop: Cannot proceed to packet capture - failed to initialize bpf device\n");
		return;
	}
	
	// Capture interface is now open, go into read() loop
	
	capend = 0;
	pktcnt = 0;
	
	while (capend == 0)
	{
		int flag;
		
		// Read next packet from bpf device
		
		flag = read(bpfd, rawfrm, BPF_READBUF);
		
		// First check for the most common event - a packet
		// has arrived.
		
		if (flag >= BPF_TRUNCATELEN)
		{
      		// Locate bpf generic header
     		frame = (struct bpf_hdr *)rawfrm;
         											
     		// Locate start of actual Ethernet frame
     		pkt = rawfrm + (frame->bh_hdrlen);		
         									
     		// Fill the pkt_record

     		// IP src addr is three 'words', or 12 bytes, from beginning
     		// of IP header - i.e. 26 bytes from start of pkt[]

     		memcpy(&(prec.srcaddr.s_addr), &(pkt[26]), 4);
     		memcpy(&(prec.dstaddr.s_addr), &(pkt[30]), 4);
     		prec.srcport = (pkt[34] & 0xFF)<<8  |  pkt[35];
     		prec.dstport = (pkt[36] & 0xFF)<<8  |  pkt[37];

     		// Copy over the length & timestamp

     		prec.len = frame->bh_datalen - 14; // IP packet is 14 bytes shorter than Ethernet frame
     		prec.time_sec = (frame->bh_tstamp).tv_sec;
     		prec.time_usec = (frame->bh_tstamp).tv_usec;

     		// Call the statistics routine with this 'event'

     		gts_add_pkt_record(gs, &prec);
         		
			// Initiate a checkpoint if the flag is set, or there's been
			// at least gs->chkpoint_interval seconds since last checkpoint (the flag
			// is most likely set as a consequence of an external user signal)

			if ((gs->last_time_sec - gs->chkpoint_lasttime) >= gs->chkpoint_interval)
				gts_checkpoint_stats(gs, CHKP_INTERIM);			
			else if (gs->chkpoint_flag == CHKP_INTERIM)
				gts_checkpoint_stats(gs, CHKP_INTERIM);
	       	else if (gs->chkpoint_flag == CHKP_FINAL)
        	{
        		// Setting capend will cause this while() loop to exit,
        		// and result in the program performing a final checkpoint
        		// and quitting gracefully.
    	    	capend = 1;
        	}
		
      		pktcnt++;
     		if ((gs->total_pkts > 0) && (pktcnt > gs->total_pkts))
     			capend = 1;
		}
		else if (flag == 0)
		{
			// We reach here if read() returned due to the timeout
			// Use this as a heart-beat so we can catch checkpoint requests
			// sent from outside the running process (e.g. user signal or
			// timer event).
			
        	if (gs->chkpoint_flag == CHKP_INTERIM)
        	{
        		gts_checkpoint_stats(gs, CHKP_INTERIM);
        	}
        	else if (gs->chkpoint_flag == CHKP_FINAL)
        	{
        		// Setting capend will cause this while() loop to exit,
        		// and result in the program performing a final checkpoint
        		// and quitting gracefully.
    	    	capend = 1;
        	}
		}
		else if (flag == -1)
		{
			if (errno != EINTR)
			{
				// Print error message only if this is NOT a case where the
				// read() call was interrupted by an external signal
				fprintf(stderr, "do_packetsnoop: error reading capture interface: ");
				perror(NULL);
			}
		}
		else
		{
			fprintf(stderr, "do_packetsnoop: read() returned insufficient (%d) bytes\n",
				flag);
				
			// Flush the current packet capture buffer to resynch
			// read() with packet stream?
			
			if (ioctl(bpfd, BIOCFLUSH, NULL) < 0)
				printf("ioctl(BIOCFLUSH) failed.\n");
		}	
	}
	
	close(bpfd);

	return;
}


//
// init_bpf()
//
// Open the BPF device, attach it to link
// interface specified by user in config file,
// program it to truncate packets, only pass UDP packets,
// run in promiscuous mode.
//

int init_bpf(struct global_state *gs)
{
	int tmpd;
	int index;
	int promisc;
	unsigned int bufsize;
	unsigned int linktype;
	struct timeval timeout;
	struct bpf_version vers;
	struct ifreq intf;
	char bpffile[128];
	struct bpf_program prgrm;
	struct bpf_insn insns[] = {
			// If not Ethernet_IP packet, jump to "exit with no packet"
			BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 12),
			BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ETHERTYPE_IP, 0, 3),
			// If not UDP, jump to "exit with no packet"
			BPF_STMT(BPF_LD+BPF_B+BPF_ABS, 23),
			BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, IPPROTO_UDP, 0, 1),
			// Getting to here means IP+UDP, so exit with truncated packet
			BPF_STMT(BPF_RET+BPF_K, BPF_TRUNCATELEN),
			BPF_STMT(BPF_RET+BPF_K, 0),				// Exit with no packet
	};
	
	
	index = 0;
	while (index < 4)	// Only search through bpf0 to bpf3
	{
		sprintf(bpffile, "%s%d", BPFDEFAULT, index);
		if ((tmpd = open(bpffile, O_RDONLY)) > 0)
			break;
		
		if (errno != EBUSY)
		{
			fprintf(stderr, "Attempt to open %s failed: ",
				bpffile);
			perror(NULL);
			return(-1);	// Return error if we failed to open
						// the BPF device for any reason other than
						// it being busy
		}	
		index++;
	}

	if (index == 4)
	{
		printf("Failed to open any bpf file\n");
		return(-1);
	}	
		
	printf("BPF device %s opened.\n", bpffile);
	
	if (ioctl(tmpd, BIOCVERSION, &vers) >= 0)
		printf("BPF version %d.%d\n", vers.bv_major, vers.bv_minor);
	else
		printf("ioctl(BIOCVERSION) failed.\n");

	timeout.tv_sec = BPF_READTIMEOUT;
	timeout.tv_usec = 0;
	if (ioctl(tmpd, BIOCSRTIMEOUT, &timeout) >= 0)
		printf("Set timeout to %d sec, %d usec\n",
		timeout.tv_sec, timeout.tv_usec);
	else
		printf("ioctl(BIOCSRTIMEOUT) failed.\n");
		
	bufsize = BPF_READBUF;
	if (ioctl(tmpd, BIOCSBLEN, &bufsize) < 0)
		printf("ioctl(BIOCSBLEN) failed.\n");
	
	bufsize = 0;
	if (ioctl(tmpd, BIOCGBLEN, &bufsize) >= 0)
		printf("Set read buffer size to %d bytes\n", bufsize);
	else
		printf("ioctl(BIOCGBLEN) failed.\n");
		
	
	memset((char *)&intf, 0, sizeof(struct ifreq));
	strcpy(intf.ifr_name, gs->snooplink);
	if (ioctl(tmpd, BIOCSETIF, &intf) >= 0)
		printf("Link interface set successfully\n");
	else
	{
		fprintf(stderr, "ioctl(BIOCSETIF) for %s failed: ", gs->snooplink);
		perror(NULL);
		close(tmpd);
		return(-1);
	}
	
	if (ioctl(tmpd, BIOCGDLT, &linktype) >= 0)
		printf("Link interface is of type %d\n",
		linktype);
	else
		printf("ioctl(BIOCGDLT) failed.\n");
	
	promisc = 0;
	if (FLAG_UNSET(gs->stats, NOPROMISC))
	{
    	if (ioctl(tmpd, BIOCPROMISC, 0) >= 0)
    		printf("BPF interface set to promiscuous mode\n");
    	else
    		promisc = 1;
	}
	
	// Program the filter engine to simply truncate all ethernet
	// packets to 38 bytes (Ether+IP+UDPports)
	
	prgrm.bf_len = sizeof(insns)/sizeof(struct bpf_insn);
	prgrm.bf_insns = insns;
	
	if (ioctl(tmpd, BIOCSETF, &prgrm) >= 0)
		printf("BPF interface set to truncate at %d bytes\n", BPF_TRUNCATELEN);
	else
	{
		fprintf(stderr, "ioctl(BIOCSETF) failed: ");
		perror(NULL);
		close(tmpd);
		return(-1);
	}
	
	// Log basic information about the packet capture interface
	
	fprintf(gs->outf, "\nPacket capture on interface %s through device %s\n",
		gs->snooplink, bpffile);
	fprintf(gs->outf, "Using BPF version %d.%d\n", vers.bv_major, vers.bv_minor);
	
	if (FLAG_SET(gs->stats, NOPROMISC) || (promisc == 1))
	{
   		printf("BPF interface NOT set to promiscuous mode\n");
   		printf("(Only traffic to/from this host will be captured)\n");
   		fprintf(gs->outf, "BPF interface NOT set to promiscuous mode\n");
   		fprintf(gs->outf, "(Only traffic to/from this host will be captured)\n");	
	}
	
	fflush(gs->outf);
	
	return(tmpd);
}


//
// init_sighandler()
//
// When running in realtime capture mode, need to install a
// mechanism for external signals to trigger checkpoints and
// initiate a graceful shutdown (final checkpoint and exit)
//
// Basic approach is to set a flag that will be checked
// by the main loop in initiate_packetsnoop() each time
// the read() call returns.
//
// Let SIGUSR1 trigger an out-of-cycle checkpoint
// Let SIGTERM trigger graceful shutdown (final checkpoint)
//

void init_sighandler(struct global_state *gs)
{
	struct sigaction tmpaction;
	
	
	memset((char *)&tmpaction, 0, sizeof(struct sigaction));
	
	tmpaction.sa_handler = usr1_sighandler;

	if (sigaction(SIGUSR1, &tmpaction, NULL) < 0)
		printf("USR1 sigaction failed\n");
	else
		printf("USR1 sigaction succeeded\n");
	
	gs_global = gs;
		
	tmpaction.sa_handler = term_sighandler;

	if (sigaction(SIGTERM, &tmpaction, NULL) < 0)
		printf("TERM sigaction failed\n");
	else
		printf("TERM sigaction succeeded\n");
	
	return;

}

//
// usr1_sighandler()
//
// Called when SIGUSR1 is trapped. Only function
// is to set the checkpoint flag so that an INTERIM
// checkpoint is executed within the next 5 seconds.
//

void usr1_sighandler(int x)
{
	// parameter x is the signal number (SIGUSR1)

	gs_global->chkpoint_flag = CHKP_INTERIM;

	return;
}


//
// term_sighandler()
//
// Called when SIGTERM is trapped. Only function
// is to set the checkpoint flag so that a FINAL
// checkpoint is executed within the next 5 seconds,
// and the program exits cleanly.
//

void term_sighandler(int x)
{
	// parameter x is the signal number (SIGTERM)

	gs_global->chkpoint_flag = CHKP_FINAL;

	return;
}




#endif WIN32


