//
// gt_printstats.cpp
//
// Routines to dump the contents of internal statstics
// structures
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with pkthisto in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


#include "stdafx.h"
#include "gt.h"


//
// gts_checkpoint_stats()
//
// Walk through the seen_flow entries, and dump to file
// all the histograms (if any) that have been created and
// stored to date. The *current* histos associated with each
// flow are NOT dumped if mode == CHKP_INTERIM, and ARE dumped
// if mode == CHKP_FINAL.
//
// Each completed histo dumped to file is then deleted from
// the seen_flow's histo lists. After a checkpoint_stats()
// call, every seen_flow entry should have no more than
// one length and one inter-arrival histo hanging off it
// (or if mode == CHKP_FINAL, no histos left at all)
//
// First time checkpoint_stats() is called we overwrite any
// existing logfiles of same names. Subsequent calls append
// data to the associated logfiles.
//
// A seen_flow entry is skipped if it hasn't seen any packets
// since the last time it was checkpoint'd.
//

void gts_checkpoint_stats(struct global_state *gs,
					  int mode)
{
	struct seen_flow *seenf;
	struct seen_servers *seens;
	int totalflows, flowsdumped;

    gs->chkpoint_flag = 0;		// Regardless of why we're here, reset the flag
	gs->chkpoint++;
	
	// If we're doing real-time packet snooping, then set the
	// checkpoint time to system time, otherwise set it based
	// on the most recent packet timestamp received. (The difference
	// is because during real-time packet capture we might have
	// checkpoints triggered by an external, user-initiated signal
	// to this process. Need to be sure the logfile shows the
	// actual system time of the checkpoint, rather than the last time
	// we received a packet.)
	
	if (FLAG_SET(gs->stats, PKTSNOOP))
		gs->chkpoint_lasttime = time(NULL);
	else
		gs->chkpoint_lasttime = gs->last_time_sec;
		
	if (mode == CHKP_FINAL)
	{
     	fprintf(gs->outf, "\n**FinalCheckpoint %d, %s\n",
     		gs->chkpoint,
     		ctime(&(gs->chkpoint_lasttime)));
    }
    else
	{
     	fprintf(gs->outf, "\n**Checkpoint %d, [%d,%d] %s\n",
     		gs->chkpoint, gs->num_flows, gs->freed_f,
     		ctime(&(gs->chkpoint_lasttime)));
    }

	totalflows = flowsdumped = 0;

	seenf = gs->sflow;
	while (seenf != NULL)
	{
		totalflows++;
		if (FLAG_SET(seenf->state, F_ESTAB))
		{
			flowsdumped++;
			gts_checkpoint_seenf(gs, seenf, mode);
			seenf->state |= F_FILEOPEN;	// Mark the flow as having had its logfiles
										// opened at least once.
		}
		seenf = seenf->next;
	}

	seens = gs->ssrv;
	while (seens != NULL)
	{
		if (seens->toserv != NULL)
		{
			if (FLAG_SET(seens->toserv->state, F_ESTAB))
			{
				gts_checkpoint_seenf(gs, seens->toserv, mode);
				seens->toserv->state |= F_FILEOPEN;	// Mark the flow as having had its logfiles
													// opened at least once.
			}
		}

		if (seens->frmserv != NULL)
		{
			if (FLAG_SET(seens->frmserv->state, F_ESTAB))
			{
				gts_checkpoint_seenf(gs, seens->frmserv, mode);
				seens->frmserv->state |= F_FILEOPEN;	// Mark the flow as having had its logfiles
														// opened at least once.
			}
		}

		seens = seens->next;
	}

	if (mode == CHKP_FINAL)
	{
		fprintf(gs->outf, "\n**Checkpointing complete.\n");
		fprintf(gs->outf, "\nHistograms were created for %d active flows (%d seen, %d removed, %d stored)\n",
			   flowsdumped, gs->num_flows, gs->freed_f, totalflows);
	}
	else
	{
		fprintf(gs->outf, "\n**EndCheckpoint %d [%d dumped, %d stored]\n",
			   gs->chkpoint, flowsdumped, totalflows);
	}
	

	fflush(gs->outf);
	return;
}


//
// gts_checkpoint_seenf()
//
// Dumps and deletes histos of a specified
// seen_flow entry.
//
// If the mode == CHKP_INTERIM then the current len and
// inter-arrival histos are left untouched.
// If the mode == CHKP_FINAL then all histos are processed.
//

void gts_checkpoint_seenf(struct global_state *gs,
					  struct seen_flow *seenf,
					  int mode)
{
	FILE *flh, *fclh, *fr, *fp, *fsl, *fsm, *fsu, *fsr;
	FILE *fih, *fcih;
	char filename[128];
	char omode[4];
	struct len_histo *tmp_lenh;
	struct len_histo **curr_lenh;
	struct inter_histo *tmp_interh;
	struct inter_histo **curr_interh;


	// It is possible to be called even though there
	// are no histos at all. Return.
	
	if (seenf->lhisto == NULL)
		return;
	
	// Set pointers to the len and inter-arrival histos
	// appropriate for CHKP_FINAL or CHKP_INTERIM modes
	
	if (mode == CHKP_FINAL)
	{
		curr_lenh = &(seenf->lhisto);
		curr_interh = &(seenf->thisto);
	}
	else
	{
		curr_lenh = &(seenf->lhisto->next);
		curr_interh = &(seenf->thisto->next);
		
		// It is possible for there to be no subsidiary
		// histos to dump, so return without needlessly
		// opening/closing files for this seen_flow entry
		
		if (seenf->lhisto->next == NULL)
			return;
	}
		
	// If this flow has already had its summary files opened,
	// we re-open in "append" mode, otherwise open in "write"
	// mode to ensure any previous files of the same name are
	// deleted.
			
	if (FLAG_SET(seenf->state, F_FILEOPEN))
		sprintf(omode, "a+");
	else
		sprintf(omode, "w");

	// Create/re-open file for this flow's histogram summary information

	if ((seenf->flow_outf = fopen(seenf->flow_outf_name, omode)) == NULL)
	{
		fprintf(gs->outf, "gts_checkpoint_seenf: opening new file %s failed, no stats printed.\n",
			seenf->flow_outf_name);
		return;
	}

	// If this is the first time, log the addr/ports and start
	// time of this flow

	if (FLAG_UNSET(seenf->state, F_FILEOPEN))
	{
		long time1 = seenf->start_sec + (seenf->start_usec/1000000);
		fprintf(seenf->flow_outf, "Src: %s:%d\nDst: %s:%d\n",
			seenf->saddr_txt, seenf->srcport,
			seenf->daddr_txt, seenf->dstport);
		fprintf(seenf->flow_outf, "Start: %s", ctime((time_t *)&time1));
	}

	fprintf(seenf->flow_outf, "Checkpoint: %d start %s",
		gs->chkpoint, ctime(&(gs->chkpoint_lasttime)));

	// Open the various output logfiles with appropriate mode

	// Length histogram ouput files start with "LH"

	sprintf(filename, "LH-%s\0", seenf->flow_outf_name);
	if ((flh = fopen(filename, omode)) == NULL)
	{
		fprintf(gs->outf, "print_flow_stats: opening new file %s failed, no stats printed.\n",
			filename);
		return;
	}

	// Length histogram ouput files start with "CLH"

	sprintf(filename, "CLH-%s\0", seenf->flow_outf_name);
	if ((fclh = fopen(filename, omode)) == NULL)
	{
		fprintf(gs->outf, "print_flow_stats: opening new file %s failed, aborting.\n",
			filename);
		exit(-1);
	}

	// Inter-arrival histograms ouput files start with "IH"

	sprintf(filename, "IH-%s\0", seenf->flow_outf_name);
	if ((fih = fopen(filename, omode)) == NULL)
	{
		fprintf(gs->outf, "print_flow_stats: opening new file %s failed, aborting.\n",
			filename);
		exit(-1);
	}

	// Cumulative inter-arrival histograms ouput files start with "CIH"

	sprintf(filename, "CIH-%s\0", seenf->flow_outf_name);
	if ((fcih = fopen(filename, omode)) == NULL)
	{
		fprintf(gs->outf, "print_flow_stats: opening new file %s failed, aborting.\n",
			filename);
		exit(-1);
	}

	sprintf(filename, "RATE-%s\0", seenf->flow_outf_name);
	if ((fr = fopen(filename, omode)) == NULL)
	{
		fprintf(gs->outf, "print_flow_stats: opening new file %s failed, aborting.\n",
			filename);
		exit(-1);
	}

	sprintf(filename, "PPS-%s\0", seenf->flow_outf_name);
	if ((fp = fopen(filename, omode)) == NULL)
	{
		fprintf(gs->outf, "print_flow_stats: opening new file %s failed, aborting.\n",
			filename);
		exit(-1);
	}

	// Only open and update SIZE* files if requested by user
	if (FLAG_SET(gs->stats, DOSIZES))
	{
     	sprintf(filename, "SIZEL-%s\0", seenf->flow_outf_name);
     	if ((fsl = fopen(filename, omode)) == NULL)
     	{
     		fprintf(gs->outf, "print_flow_stats: opening new file %s failed, aborting.\n",
     			filename);
     		exit(-1);
     	}

     	sprintf(filename, "SIZEM-%s\0", seenf->flow_outf_name);
     	if ((fsm = fopen(filename, omode)) == NULL)
     	{
     		fprintf(gs->outf, "print_flow_stats: opening new file %s failed, aborting.\n",
     			filename);
     		exit(-1);
     	}

     	sprintf(filename, "SIZEU-%s\0", seenf->flow_outf_name);
     	if ((fsu = fopen(filename, omode)) == NULL)
     	{
     		fprintf(gs->outf, "print_flow_stats: opening new file %s failed, aborting.\n",
     			filename);
     		exit(-1);
     	}

     	sprintf(filename, "SIZER-%s\0", seenf->flow_outf_name);
     	if ((fsr = fopen(filename, omode)) == NULL)
     	{
     		fprintf(gs->outf, "print_flow_stats: opening new file %s failed, aborting.\n",
     			filename);
     		exit(-1);
     	}
	}

	// Sort the current set of histos into ascending order of
	// start times rather than descending order.

	gt_sort_flow_histos(gs, seenf, mode);

	// Begin stepping through the histos

	while (*curr_lenh != NULL)
	{
		tmp_lenh = *curr_lenh;
		tmp_interh = *curr_interh;

		if (FLAG_SET(tmp_lenh->state, F_ESTAB))
			seenf->time_established +=
				(long)((tmp_lenh->endtime - tmp_lenh->starttime)/1000000);

		if (tmp_lenh->elements > gs->pkt_cnt_min)
		{	
			int lowerlen = -1;
			int upperlen = -1;
			int midlen = -1;
			double rate = (8*tmp_lenh->totallen) /
				((float)(tmp_lenh->endtime - tmp_lenh->starttime) / 1000);

			long starttime = seenf->start_sec +
				(long)((seenf->start_usec + tmp_lenh->starttime)/1000000);


			// Calculate and output the length histo statistics
			
			if (FLAG_SET(gs->stats, COMPRESS_ASCII))
				gts_compressed_lenhisto(gs, tmp_lenh, flh, fclh, &lowerlen, &midlen, &upperlen);
			else
				gts_ascii_lenhisto(gs, tmp_lenh, flh, fclh, &lowerlen, &midlen, &upperlen);

			fprintf(seenf->flow_outf,
				"Histo:%03d: %6.3f - %6.3f min, avg %3.1f bytes, %5.2f kbps, %5.2f pps,",
				tmp_lenh->histo_num, tmp_lenh->starttime/(float)60000000,
				tmp_lenh->endtime/(float)60000000,
				tmp_lenh->totallen/(float)tmp_lenh->elements, rate,
				tmp_lenh->elements / ((float)(tmp_lenh->endtime - tmp_lenh->starttime) / 1000000)
				);

			fprintf(seenf->flow_outf,
				" l/m/h %d/%d/%d, %d pkts, %d err\n",
				lowerlen, midlen, upperlen,
				tmp_lenh->elements, tmp_lenh->errored_cnts);

			fprintf(fr, "%d %5.2f\n",
				starttime,
				rate);

			fprintf(fp, "%d %5.2f\n",
				starttime,
				tmp_lenh->elements / ((float)(tmp_lenh->endtime - tmp_lenh->starttime) / 1000000)
				);

			// Only output SIZE* data to unique files if requested by user
			if (FLAG_SET(gs->stats, DOSIZES))
			{
    			fprintf(fsm, "%d %d\n",
    				starttime,
    				midlen);

    			fprintf(fsl, "%d %d\n",
    				starttime,
    				lowerlen);

    			fprintf(fsu, "%d %d\n",
    				starttime,
    				upperlen);

    			if ((midlen-lowerlen) >= 1)
    				fprintf(fsr, "%d %6.3f\n",
    					starttime,
    					(upperlen-midlen)/(float)(midlen-lowerlen));
			}


			// Calculate and output the inter-arrival histo statistics

			if (FLAG_SET(gs->stats, COMPRESS_ASCII))
				gts_compressed_interhisto(gs, tmp_interh, fih, fcih);
			else
				gts_ascii_interhisto(gs, tmp_interh, fih, fcih);

		}

		// Now delete the current histos and move to the
		// next ones (if any)

		*curr_lenh = tmp_lenh->next;
		*curr_interh = tmp_interh->next;
		free(tmp_lenh);
		free(tmp_interh);
	}

	fprintf(seenf->flow_outf, "Checkpoint: %d end\n", gs->chkpoint);
	
	if (mode == CHKP_FINAL)
	{
		long time1 = seenf->start_sec + (seenf->start_usec/1000000);

		fprintf(gs->outf, "%s: %s:%d -> %s:%d: %d elements, %6.3f min, started %s",
				seenf->flow_outf_name,
				seenf->saddr_txt, seenf->srcport,
				seenf->daddr_txt, seenf->dstport,
				seenf->num_pkts,
				seenf->time_established/(float)60,
				ctime((time_t *)&time1));

		fprintf(seenf->flow_outf, "TotalPackets: %d\nTotalTimeInEstablished: %6.3f min\n",
			seenf->num_pkts,
			seenf->time_established/(float)60);
	}
	else
	{
		fprintf(gs->outf, "%s: %s:%d -> %s:%d: %d elements, %6.3f min\n",
				seenf->flow_outf_name,
				seenf->saddr_txt, seenf->srcport,
				seenf->daddr_txt, seenf->dstport,
				seenf->num_pkts,
				seenf->time_established/(float)60);
	}


	// Close the various logfiles

	fclose(flh);
	fclose(fclh);
	fclose(fr);
	fclose(fp);
	fclose(fih);	
	fclose(fcih);
	fclose(seenf->flow_outf);
	
	// Only close SIZE* files if we previously opened them
	if (FLAG_SET(gs->stats, DOSIZES))
	{
    	fclose(fsl);
    	fclose(fsm);
    	fclose(fsu);
    	fclose(fsr);
	}

	return;
}


//
// gt_sort_flow_histos()
//
// Re-orders the len and inter-arrival histos of
// a particular seen_flow entry, in ascending order
// of start time. [The per-packet stats routines currently
// build the len_histo and inter_histo lists in reverse
// order so that the most recent histo is at the head of
// the list. This needs to be reversed prior to printing
// when checkpointing is activated.]
//
// If mode == CHKP_INTERIM, we re-order histos *below*
// the current histo. If mode == CHKP_FINAL, we re-order
// ALL histos.
//

void gt_sort_flow_histos(struct global_state *gs,
					struct seen_flow *seenf,
					int mode)
{

	struct len_histo *tmplh, *tmplh2, *prevlh;
	struct inter_histo *tmpih, *tmpih2, *previh;
	struct len_histo **curr_lenh;
	struct inter_histo **curr_interh;

	if (mode == CHKP_FINAL)
	{
		curr_lenh = &(seenf->lhisto);
		curr_interh = &(seenf->thisto);
	}
	else
	{
		curr_lenh = &(seenf->lhisto->next);
		curr_interh = &(seenf->thisto->next);
	}

	// Simply reverse the order of the len histo list

	prevlh = NULL;			// Top element on new list
	tmplh = *curr_lenh;		// Top element of old list to be moved over 



	while (tmplh != NULL)
	{
		tmplh2 = tmplh->next;	// Grab pointer to next element on old list

		tmplh->next = prevlh;	// Take element to be moved over, link it
								// above latest element on new list

		prevlh = tmplh;			// Moved element now becomes top element of
								// new list.
		tmplh = tmplh2;			// 2nd element on old list now becomes top of old list

	}

	// At this point prevlh points to the top element of the
	// newly re-ordered list.

	*curr_lenh = prevlh;

	// Repeat the process for the inter-arrival hist list

	previh = NULL;					// Top element on new list
	tmpih = *curr_interh;	// Top element of old list to be moved over 


	while (tmpih != NULL)
	{
		tmpih2 = tmpih->next;	// Grab pointer to next element on old list

		tmpih->next = previh;	// Take element to be moved over, link it
								// above latest element on new list

		previh = tmpih;			// Moved element now becomes top element of
								// new list.
		tmpih = tmpih2;			// 2nd element on old list now becomes top of old list

	}

	// At this point previh points to the top element of the
	// newly re-ordered list.

	*curr_interh = previh;

	return;
}

//
// gts_ascii_lenhisto()
//
// Dump a single length histogram to file, and the
// cumulative distribution, in pure ASCII "X Y" pair
// format.
//

void gts_ascii_lenhisto(struct global_state *gs,
						struct len_histo *tmp_lenh,
						FILE *flh,
						FILE *fclh,
						int *low,
						int *mid,
						int *hi)
{

	int tmplen;
	long totalcount = 0;
	float fraction;

   	fprintf(flh, "\n\"LH %d\n", tmp_lenh->histo_num);
   	fprintf(fclh, "\n\"CLH %d\n", tmp_lenh->histo_num);
			
   	tmplen = tmp_lenh->smallestelement;
   	while (tmplen <= tmp_lenh->largestelement)
   	{
   		fprintf(flh, "%d %5.2f\n",
   			tmplen*gs->length_histo_width,
   			100*(tmp_lenh->counts[tmplen])/(float)tmp_lenh->elements);

   		totalcount += tmp_lenh->counts[tmplen];
   		fraction = 100*(totalcount)/(float)tmp_lenh->elements;

   		if ((fraction >= (float)5.0) && (*low == -1))
   			*low = tmplen*gs->length_histo_width;

   		if ((fraction >= (float)50.0) && (*mid == -1))
   			*mid = tmplen*gs->length_histo_width;

   		if ((fraction >= (float)95.0) && (*hi == -1))
   			*hi = tmplen*gs->length_histo_width;

   		// Only dump cumulative plot data if we're talking about
   		// points betwen 0.5% and 99.5%, otherwise this is pointless
   		// information (and just takes up loads of disk space on
   		// long running real time captures)
				
   		if ((fraction > (float)0.5) && (fraction < (float)99.5))
   			fprintf(fclh, "%d %5.2f\n",
   				tmplen*gs->length_histo_width, fraction);
						
   		tmplen++;
   	}

   	if (*mid == -1)
   		*mid = tmplen*gs->length_histo_width;

   	if (*hi == -1)
   		*hi = tmplen*gs->length_histo_width;

   	return;
}

//
// gts_compressed_lenhisto()
//
// Dump a single length histogram to file, and the
// cumulative distribution, in compressed ASCII
// format.

void gts_compressed_lenhisto(struct global_state *gs,
						struct len_histo *tmp_lenh,
						FILE *flh,
						FILE *fclh,
						int *low,
						int *mid,
						int *hi)
{

	int tmplen;
	long totalcount;
	float fraction;
	char tmpline[2048];	// Two characters per bucket, so this is enough for
						// roughly 1020 buckets (+ some bytes).
	int sindex;

	// Start with lame and probably unecessary sanity check.
		
	if ((tmp_lenh->largestelement - tmp_lenh->smallestelement) > 1000)
	{
		printf("pkthisto: WARNING, too many elements in length histograms.\n");
		printf("Need to update gts_compressed_lenhisto() and recompile pkthisto.\n");
	}
	
   	fprintf(flh, "LH %d %d %d %d %d ",
   				tmp_lenh->histo_num,
   				gs->length_histo_width,
   				tmp_lenh->smallestelement,
   				tmp_lenh->largestelement,
   				tmp_lenh->errored_cnts);
   				
   	tmpline[0] = '>';
   	sindex = 1;
   	
   	tmplen = tmp_lenh->smallestelement;
   	while (tmplen <= tmp_lenh->largestelement)
   	{
        int elem = tmp_lenh->counts[tmplen];
        int nxtlen = tmplen;
        int diff;

        tmpline[sindex] = (elem/64) + 33;
        tmpline[sindex+1] = (elem%64) + 33;
        sindex += 2;

        while ((tmp_lenh->counts[nxtlen+1] == elem) &&
                        (nxtlen < tmp_lenh->largestelement))
                nxtlen++;

        // If there's more than two buckets with same
        // value as the bucket whose value we just 'printed'
        // then add "znn%" to the output stream (where nn is
        // the number of times the preceding value ought
        // to be repeated)

        if ((diff = (nxtlen - tmplen)) > 2)
        {
                char dummy[16];
                sprintf(dummy, "z%i%%\0", diff);
                memcpy(&(tmpline[sindex]), dummy, strlen(dummy));
                sindex += strlen(dummy);
                tmplen += diff;
        }

  		tmplen++;
   	}
   	
   	tmpline[sindex] = 0;
   	fprintf(flh, "%s\n", tmpline);

   	
   	// Generate the cumulative distribution, and
   	// calculate the 5%, 50%, and 95% lengths.
   	
   	fprintf(fclh, "CLH %d %d %d %d ",
   				tmp_lenh->histo_num,
   				gs->length_histo_width,
   				tmp_lenh->smallestelement,
   				tmp_lenh->largestelement);
   	
   	tmpline[0] = '>';
   	sindex = 1;
   	totalcount = 0;
   	
   	tmplen = tmp_lenh->smallestelement;
   	while (tmplen <= tmp_lenh->largestelement)
   	{
   		int elem;
  		
   		totalcount += tmp_lenh->counts[tmplen];
   		elem = (4096*totalcount)/tmp_lenh->elements;

        tmpline[sindex] = (elem/64) + 33;
        tmpline[sindex+1] = (elem%64) + 33;
        sindex += 2;

   		fraction = 100*(totalcount)/(float)tmp_lenh->elements;

   		if ((fraction >= (float)5.0) && (*low == -1))
   			*low = tmplen*gs->length_histo_width;

   		if ((fraction >= (float)50.0) && (*mid == -1))
   			*mid = tmplen*gs->length_histo_width;

   		if ((fraction >= (float)95.0) && (*hi == -1))
   			*hi = tmplen*gs->length_histo_width;
  		
  		tmplen++;
   	}
   	
   	tmpline[sindex] = 0;
   	fprintf(fclh, "%s\n", tmpline);
   		
   	if (*mid == -1)
   		*mid = tmplen*gs->length_histo_width;

   	if (*hi == -1)
   		*hi = tmplen*gs->length_histo_width;

   	return;
}




//
// gts_ascii_interhisto()
//
// Dump a single interval histogram to file, and the
// cumulative distribution, in pure ASCII "X Y" pair
// format.
//

void gts_ascii_interhisto(struct global_state *gs,
						struct inter_histo *tmp_interh,
						FILE *fih,
						FILE *fcih)
{

	int tmpintv;
	long totalinter = 0;
	float fraction;
	float scale = gs->interval_histo_width/(float)1000;

  	fprintf(fih, "\n\"IH %d\n", tmp_interh->histo_num);
  	fprintf(fcih, "\n\"CIH %d\n", tmp_interh->histo_num);

  	tmpintv = tmp_interh->smallestelement;
  	while (tmpintv <= tmp_interh->largestelement)
  	{
  		fprintf(fih, "%4.2f %5.2f\n",
  			tmpintv*scale,
  			100*(tmp_interh->counts[tmpintv])/(float)tmp_interh->elements);
					
  		totalinter += tmp_interh->counts[tmpintv];
  		fraction = 100*(totalinter)/(float)tmp_interh->elements;
				
  		// Only dump cumulative plot data if we're talking about
  		// points betwen 0.5% and 99.5%, otherwise this is pointless
  		// information (and just takes up loads of disk space on
  		// long running real time captures)
				
  		if ((fraction > (float)0.5) && (fraction < (float)99.5))
  			fprintf(fcih, "%4.2f %5.2f\n",
  				tmpintv*scale, fraction);
						
  		tmpintv++;
  	}

   	return;
}

//
// gts_cmopressed_interhisto()
//
// Dump a single interval histogram to file, and the
// cumulative distribution, in compressed ASCII
// format.

void gts_compressed_interhisto(struct global_state *gs,
						struct inter_histo *tmp_interh,
						FILE *fih,
						FILE *fcih)
{

	int tmpintv;
	long totalinter;
	char tmpline[2048];	// Two characters per bucket, so this is enough for
						// roughly 1020 buckets (+ some bytes).
	int sindex;

	// Start with lame and probably unecessary sanity check.
		
	if ((tmp_interh->largestelement - tmp_interh->smallestelement) > 1000)
	{
		printf("pkthisto: WARNING, too many elements in interval histograms.\n");
		printf("Need to update gts_compressed_interhisto() and recompile pkthisto.\n");
	}
	
   	fprintf(fih, "IH %d %d %d %d %d ",
   				tmp_interh->histo_num,
   				gs->interval_histo_width,
   				tmp_interh->smallestelement,
   				tmp_interh->largestelement,
   				tmp_interh->errored_cnts);
   				
   	tmpline[0] = '>';
   	sindex = 1;
   	
   	tmpintv = tmp_interh->smallestelement;
   	while (tmpintv <= tmp_interh->largestelement)
   	{
        int elem = tmp_interh->counts[tmpintv];
        int nxtintv = tmpintv;
        int diff;

        tmpline[sindex] = (elem/64) + 33;
        tmpline[sindex+1] = (elem%64) + 33;
        sindex += 2;

        while ((tmp_interh->counts[nxtintv+1] == elem) &&
                        (nxtintv < tmp_interh->largestelement))
                nxtintv++;

        // If there's more than two buckets with same
        // value as the bucket whose value we just 'printed'
        // then add "+nn%" to the output stream (where nn is
        // the number of times the preceding value ought
        // to be repeated)

        if ((diff = (nxtintv - tmpintv)) > 2)
        {
                char dummy[16];
                sprintf(dummy, "z%i%%\0", diff);
                memcpy(&(tmpline[sindex]), dummy, strlen(dummy));
                sindex += strlen(dummy);
                tmpintv += diff;
        }

  		tmpintv++;
   	}
   	
   	tmpline[sindex] = 0;
   	fprintf(fih, "%s\n", tmpline);

   	
   	// Generate the cumulative distribution, and
   	// calculate the 5%, 50%, and 95% lengths.
   	
   	fprintf(fcih, "CIH %d %d %d %d ",
   				tmp_interh->histo_num,
   				gs->interval_histo_width,
   				tmp_interh->smallestelement,
   				tmp_interh->largestelement);
   	
   	tmpline[0] = '>';
   	sindex = 1;
   	totalinter = 0;
   	
   	tmpintv = tmp_interh->smallestelement;
   	while (tmpintv <= tmp_interh->largestelement)
   	{
   		int elem;
  		
   		totalinter += tmp_interh->counts[tmpintv];
   		elem = (4096*totalinter)/tmp_interh->elements;

        tmpline[sindex] = (elem/64) + 33;
        tmpline[sindex+1] = (elem%64) + 33;
        sindex += 2;

  		tmpintv++;
   	}
   	
   	tmpline[sindex] = 0;
   	fprintf(fcih, "%s\n", tmpline);
   		
  	return;
}








