//
// gt_startup.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with pkthisto in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

#include "stdafx.h"
#include "gt.h"


//
// parse_config_file()
//
// Parses the config file to set all parameters
// controlling the subsequent tracefile analysis.
//
//
void parse_config_file(struct global_state *gs,
					   char *configfile)
{
	char confstr[128];	// Store for each line of config file
	FILE *conf = NULL;	// File handle for config file
	int confeof;

	// Open and parse the config file

	conf = fopen(configfile, "rb");
	if (conf == NULL)
	{
		printf("Opening config file %s failed.\n",
			configfile);
		return;
	}

	confeof = 0;
	while (confeof != 1)
	{
		memset(confstr, 0, 128);
		if (fgets(confstr, 126, conf) == NULL)
			 confeof = 1;
		else
		{			
			if (confstr[0] != '#')
				parse_config_line(gs, confstr);
		}
	}
}

//
// parse_config_line()
//
// Parses one line of the config file to set parameters
// controlling the subsequent trace file analysis.
//
void parse_config_line(struct global_state *gs,
					   char *line)
{
	char	c1[128];	// First config line param
	char	c2[128];	// Second config line param
	char	c3[128];	// Third config line param
	char	*cspace;	// Pointer to each word in config line

	if (sscanf(line, "%s", c1) <= 0)
		return;

	memset(c2, 0, 128);
	memset(c3, 0, 128);

	cspace = strchr(line, ' ');
	if (cspace != NULL)
	{
		sscanf(cspace, "%s", c2);
		cspace = strchr(cspace+strlen(c2), ' ');
		if (cspace != NULL)
			sscanf(cspace, "%s", c3);
	}

	if (strcmp(c1, "sourcefile\0") == 0)
	{	// Specify the filename for a single source tracefile
		strcpy(gs->tracefile, c2);
	}

	else if (strcmp(c1, "outfile_prefix\0") == 0)
	{	// Specify the common prefix for output filenames
		strcpy(gs->outf_prefix, c2);
	}

	else if (strcmp(c1, "file_is_tcpdump\0") == 0)
	{	// Trace file is in tcpdump format
		gs->stats |= TCPDUMP;
	}

	else if (strcmp(c1, "file_is_sniffer\0") == 0)
	{	// Trace file is in NAI Sniffer format
		gs->stats |= SNIFFER;
	}
	
	else if (strcmp(c1, "realtime_capture\0") == 0)
	{
		// Take trace input from realtime capture off
		// a specified local interface
		gs->stats |= PKTSNOOP;
		strcpy(gs->snooplink, c2);
	}

	else if (strcmp(c1, "no_promiscuous\0") == 0)
	{
		// When doing real-time capture, do not put the
		// capture interface into promiscuous mode (the
		// default is to attempt promiscuous mode)
		gs->stats |= NOPROMISC;
	}

	else if (strcmp(c1, "compress_output\0") == 0)
	{
		// Instead of dumping each histgram as a series of
		// "X Y" pairs suitable for xgraph, use a compressed
		// format where each histogram is represented by a single
		// line of ASCII characters.
		gs->stats |= COMPRESS_ASCII;
	}

	else if (strcmp(c1, "dump_sizes\0") == 0)
	{
		// When creating output files, do specific files for
		// low/mid/upper packet lengths per histogram (although this
		// can be extracted in post-processing from the 'gtout*.txt'
		// file). By default this is not done (to save file system
		// space)
	
		gs->stats |= DOSIZES;
	}

	else if (strcmp(c1, "min_pkts_in_flow\0") == 0)
	{	// This number of packets must be received for a flow
		// before we bother creating an output file to dump
		// histogram information
		 gs->pkt_cnt_min = strtol(c2,(char **)NULL,0);
		 if (gs->pkt_cnt_min < 10)
		 {
			 printf("Config error: 'min_pkts_in_flow %d' too low, setting to 100\n",
				 gs->pkt_cnt_min);
			 gs->pkt_cnt_min = 100;
		 }
	}

	else if (strcmp(c1, "max_pkts_per_histo\0") == 0)
	{	// This is the maximum number of packets per histogram
		// in a given flow. A new histogram will be created each
		// time this value is exceeded.
		 gs->pkt_per_histo = strtol(c2,(char **)NULL,0);
		 if (gs->pkt_per_histo < 200)
		 {
			 printf("Config error: 'min_pkts_in_flow %d' too low, setting to 200\n",
				 gs->pkt_per_histo);
			 gs->pkt_per_histo = 200;
		 }
	}

	else if (strcmp(c1, "total_pkts\0") == 0)
	{	// For real-time capture, finish after receiving
		// and processing this many packets. If this is not
		// set, the program will loop until interrupted
		// with SIGTERM.
		 gs->total_pkts = strtol(c2,(char **)NULL,0);
		 if (gs->total_pkts < 0)
		 {
			 printf("Config error: 'total_pkts %d' is invalid, setting to 0 (no limit)\n",
				 gs->total_pkts);
			 gs->total_pkts = 0;
		 }
	}

	else if (strcmp(c1, "checkpoint_interval\0") == 0)
	{
		int n;
		// Checkpoint the data sets every X minutes
		n = strtol(c2,(char **)NULL,0);
		if (n < 1)
		{
			 printf("Config error: checkpoint_interval must be > 1 min. Setting to %d min\n",
				 CHKPOINTINTERVAL/60);
			 gs->chkpoint_interval = CHKPOINTINTERVAL;
		}
		else
			 gs->chkpoint_interval = n*60;
	}

	else if (strcmp(c1, "flow_max_milliseconds\0") == 0)
	{
		long tmpgap;
		// If this number of milliseconds elapses between packets
		// in the same flow, the flow is considered to have STOPPED.
		// Minimum is MIN_FLOW_MAX_MS, default is FLOW_MAX_MS_DEFAULT
		 tmpgap = strtol(c2,(char **)NULL,0);
		 if (tmpgap < MIN_FLOW_MAX_MS)
		 {
			 printf("Config error: 'flow_max_milliseconds %d' too low, setting to %dms\n",
				 tmpgap, FLOW_MAX_MS_DEFAULT);
			 tmpgap = FLOW_MAX_MS_DEFAULT;
		 }
		 gs->flow_max_gap = tmpgap*1000;
	}

	else if (strcmp(c1, "specific_server\0") == 0)
	{
		struct seen_servers *newserv;
		// Specify an address/port pair representing a known
		// server node. Two additional flows will be tracked
		// representing the aggregate of all traffic to and from
		// this server node. (This option may be specified more than
		// once if we are watching the traffic from multiple servers.)

		if ((newserv = (struct seen_servers *)malloc(sizeof(struct seen_servers)))
			== NULL)
		{
			 printf("Config error: malloc failed for specific_server, failed.\n");
		}
		else
		{
			memset(newserv, 0, sizeof(struct seen_servers));
			(newserv->srvaddr).s_addr  = inet_addr(c2);
			newserv->srvport = strtol(c3,(char **)NULL,0);
			newserv->next = gs->ssrv;
			gs->ssrv = newserv;
		}
	}

	else
	{
		printf("Config file error: %s is not a valid token.\n", c1);
		fprintf(gs->outf, "Config file error: %s is not a valid token.\n", c1);
	}

	return;
}



//
// sanity_check_config()
//
// Do some basic sanity checks on the analysis parameters
// we're about to start using. Returns 0 if okay, -1 if we
// should quit.
//

int sanity_check_config(struct global_state *gs)
{
	int err = 0;
	int n;

	if (gs->pkt_cnt_min >= gs->pkt_per_histo)
	{
		fprintf(gs->outf, "Sanity error: pkt_cnt_min(%d) >= pkt_per_histo(%d)!\n",
			gs->pkt_cnt_min, gs->pkt_per_histo);
		printf("Sanity error: pkt_cnt_min(%d) >= pkt_per_histo(%d)!\n",
			gs->pkt_cnt_min, gs->pkt_per_histo);
		err |= -1;
	}

	n = FLAG_SET(gs->stats, TCPDUMP) +
		FLAG_SET(gs->stats, SNIFFER) +
		FLAG_SET(gs->stats, PKTSNOOP);
	
	if (n == 0)
	{
		fprintf(gs->outf, "Sanity error: Must specify TCPDUMP, SNIFFER, or PKTSNOOP format\n");
		printf( "Sanity error: Must specify TCPDUMP, SNIFFER, or PKTSNOOP format\n");
		err |= -1;
	}

	if (n > 1)
	{
		fprintf(gs->outf, "Sanity error: %d Only ONE of TCPDUMP, SNIFFER, or PKTSNOOP formats\n", n);
		printf( "Sanity error: %d Only ONE of TCPDUMP, SNIFFER, or PKTSNOOP formats\n", n);
		err |= -1;
	}


	return(err);
}