//
// gt_statistics.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with pkthisto in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//
// These are the main routines for asynchronously building
// and managing the per-flow histograms.
//
// Primary entry point is:  gts_add_pkt_record()
//
// Prior to using gts_add_pkt_record() it is assumed that
// an appropriate 'struct global_state' has been established
// with initial parameters.
//
//

#include "stdafx.h"
#include "gt.h"


// gts_add_pkt_record()
//
// This is the top level entry point into the
// statistics routines. Called for every 'event'
// of a new packet record (or possibly a new packet
// arrival, if we're doing real-time analysis)
//

void gts_add_pkt_record(struct global_state *gs,
			   struct pkt_record *prec)
{
	struct seen_flow *seenf;
	unsigned int flowcntr;		// Used to calculate new flow's filename suffix
	int tofrmserv;

	gs->last_time_sec = prec->time_sec;	// Global copy of most recent timestamp

	// Find a matching flow (or create one if necessary)
	// and call the flow_update routine.

	// [For now the list of flows is a simplistic linear
	// linked list - not very efficient, will need to redesign
	// it for real-time use later.]
	//
	// 080601: Faster look-up hack: A match is moved to the head
	// of the linked list if it is more than 20 entries down the
	// list when found. This should ensure most currently active
	// flows are in the top 20 positions of the linear list.
	// (This works on the assumption that there's no more than
	// 20 concurrent and active flows of interest at any one
	// time... if there's more than 20, there'll be constant
	// churn in the linked list.)

	flowcntr = 1;
	seenf = gs->sflow;
	while (seenf != NULL)
	{
		if (((prec->dstaddr).s_addr == (seenf->dstaddr).s_addr) &&
			((prec->srcaddr).s_addr == (seenf->srcaddr).s_addr) &&
			(prec->dstport == seenf->dstport) &&
			(prec->srcport == seenf->srcport))
		{
			// Update this flow's histograms
			gts_add_flow_record(gs, prec, seenf);

			// Update the associated server's aggregate flow
			// histograms (if there is an associated server flow)
			if (seenf->server != NULL)
				gts_add_flow_record(gs, prec, seenf->server);

			// If this entry is too far down the linked list,
			// pop it to the top and move everyone else down one.
			if (flowcntr > 20)
			{
				// Need to make sure gts_check_idle_flows()
				// housekeeping isn't locked into a loop because
				// we pop this entry to the head of list. If
				// gs->nxt_chkf actually points to this entry,
				// adjust it to point to the previous entry.
				
				if (gs->nxt_chkf == seenf)
					gs->nxt_chkf = seenf->prev;
					
				// Move seenf from its current
				// position up to head of the list.
				// (Note that since we're at least 20
				// entries from the head of the list we
				// can assume seenf->prev is non-null.)

				seenf->prev->next = seenf->next;
				if (seenf->next != NULL)
					seenf->next->prev = seenf->prev;
				seenf->next = gs->sflow;
				gs->sflow->prev = seenf;
				seenf->prev = NULL;
				gs->sflow = seenf;
			}

			// Some housekeeping each time a packet event is processed.
			// Take one step through the seen_flow list, and if the next
			// seen_flow entry is idle we remove the topmost histogram
			// structs to save memory. (Only one seen_flow entry is checked
			// per packet event, spreading the load evenly over time.)
			
			gts_check_idle_flows(gs, prec->time_sec, prec->time_usec);

			return; // We're done with this packet.
		}
		flowcntr++;
		seenf = seenf->next;
	}

	// Getting to here means there's no current match in the
	// existing seen_flow list.
	
	// Create and initialize a new seen_flow entry for this flow.

	if ((seenf = (struct seen_flow *)malloc(sizeof(struct seen_flow)))
		== NULL)
	{
		fprintf(gs->outf, "add_pkt_record: malloc failed, exiting!\n");
		printf("add_pkt_record: malloc failed, exiting!\n");
		exit(-1);
	}

	// Zero all elements of this new seen_flow entry

	memset((char *)seenf, 0, sizeof(struct seen_flow));

	// Establish non-zero elements

	(seenf->dstaddr).s_addr = (prec->dstaddr).s_addr;
	(seenf->srcaddr).s_addr = (prec->srcaddr).s_addr;
	seenf->dstport = prec->dstport;
	seenf->srcport = prec->srcport;
	seenf->start_sec = prec->time_sec;
	seenf->start_usec = prec->time_usec;

	// Create ascii versions of both ip addresses for later use

	sprintf(seenf->saddr_txt, "%s", inet_ntoa(seenf->srcaddr));
	sprintf(seenf->daddr_txt, "%s", inet_ntoa(seenf->dstaddr));

	// Link this new seen_flow into the global list,
	// and if this is the first flow, initialize the
	// gs->nxt_chkf pointer too.

	seenf->next = gs->sflow;	// We point to previous head of list
	if (gs->sflow != NULL)
		gs->sflow->prev = seenf;	// Previous head points back to us
	
	if (gs->sflow == NULL)
		gs->nxt_chkf = seenf;
		
	gs->sflow = seenf;

	// Optionally note the start of this new flow

	gt_flowstart_event(gs, seenf, prec->time_sec, prec->time_usec);

	// Call the gts_add_flow_record() routine to initialize the
	// remaining fields (histograms, etc)

	gts_add_flow_record(gs, prec, seenf);

	// Check if this pkt needs to contribute to
	// any aggregate to/from server node flow statistics

	tofrmserv = gts_init_srv_flows(gs, seenf, prec);

	// Pre-assign an output filename, in case we do decide
	// to begin dumping results to file. Suffixes start at "1"
	// (since "0" is used for the overall status output messages)
	// Based on the returned value from gts_init_srv_flows() we
	// optionally add "TSn", "FSn" or nothing to the numerical
	// suffix to indicate if the flow is to/from an identified
	// server.
	
	gs->num_flows++;

	if (FLAG_SET(tofrmserv, TO_SERVER))
		sprintf(seenf->flow_outf_name, "%s%dTS%d.txt",
			gs->outf_prefix, gs->num_flows, (tofrmserv & 0x0F));

	else if (FLAG_SET(tofrmserv, FROM_SERVER))
		sprintf(seenf->flow_outf_name, "%s%dFS%d.txt",
			gs->outf_prefix, gs->num_flows, (tofrmserv & 0x0F));

	else sprintf(seenf->flow_outf_name, "%s%d.txt",
			gs->outf_prefix, gs->num_flows);

	return;
}



//
// gts_add_flow_record()
//
// Takes a record, and pointer to a seen_flow entry,
// and updates the various histograms. This routine
// is only called when an actual packet arrival event
// for this flow has occurred.
//

void gts_add_flow_record(struct global_state *gs,
					 struct pkt_record *prec,
					 struct seen_flow *seenf)
{
	int new_histo_flag = 0;
	__int64 new_relative_time;	// NB: A 'long' isn't big enough under MSVisualC++
								// (it is only 32 bits wide, same as an 'int')

	// Increment the packet counter for this flow

	seenf->num_pkts++;
	seenf->num_pkts_current++;

	// Calculate the time of this event in microseconds relative
	// to the time this flow started.

	new_relative_time = (__int64)1000000*(prec->time_sec - seenf->start_sec);
	new_relative_time += (__int64)(prec->time_usec - seenf->start_usec);

	// If there's no current length histo, this must be the first
	// pkt received for a new flow. Set flag to ensure initialization
	// of pristine length and inter-arrival histograms. Otherwise,
	// check to see if this pkt is so long after the previous pkt
	// on this flow that we consider the flow to have stopped and
	// restarted - in this case, also intialize a new histogram.

	if ((seenf->lhisto == NULL) ||
	    ((new_relative_time - seenf->last_pkt_usec) >= gs->flow_max_gap))
			new_histo_flag |= F_INIT;

	// If more than gs->pkt_per_histo packets have arrived on this
	// flow's current histogram, set a flag that will cause new
	// histograms to be started

	if (seenf->num_pkts_current >= gs->pkt_per_histo)
		new_histo_flag |= F_ESTAB;

	// If we're about to make a new set of histos, increment
	// the histo number counter

	if (new_histo_flag != 0)
		seenf->histo_num++;

	// Update the packet length histogram

	update_plen_histo(gs, prec, seenf, new_relative_time, new_histo_flag);

	// Update the inter-packet arrival time histogram

	update_arrival_histo(gs, prec, seenf, new_relative_time, new_histo_flag);

	// Update the seen_flow entry's timestamp

	seenf->last_pkt_usec = new_relative_time;

	// If the current length histogram's instance of this flow is in
	// the F_INIT state and we've seen enough pkts, move it to
	// the F_ESTAB state. (The state of the inter-arrival histo is
	// the same, so we don't adjust it explicitly.) Trigger optional
	// "flow is established" event.

	if (FLAG_SET(seenf->lhisto->state, F_INIT))
		if (seenf->lhisto->elements > gs->pkt_cnt_min)
		{
			// Call gt_flow_established() BEFORE setting
			// the seenf->state flag to F_ESTAB and resetting
			// the F_INIT flag
			gt_flow_established(gs, seenf);
			seenf->lhisto->state |= F_ESTAB;
			seenf->lhisto->state &= ~F_INIT;
			seenf->state |= F_ESTAB;
		}
}



//
// update_plen_histo()
//
// Adds a new data point to the length histogram, creates a
// new histogram if needed (flag parameter non zero), and
// deletes the previous histogram for this flow if it hadn't
// yet reached F_ESTAB state (saves memory, no point in keeping
// histogram structures around that wont be processed later
// anyway).
//
// [If it decides to remove the previous len histo, it also
// removes previous inter-arrival histo.]
//

void update_plen_histo(struct global_state *gs,
					   struct pkt_record *prec,
					   struct seen_flow *seenf,
					   __int64 newtime,
					   int flag)
{
	struct len_histo *tmphisto;
	int index;

	// If 'flag' parameter is non-zero, start a new histogram

	tmphisto = seenf->lhisto;
	if (flag > 0)
	{
		// First, if the current histo hasn't reached F_ESTAB state
		// yet, then toss it away to save space (since we're not going
		// to analyze it later anyway)

		if (seenf->lhisto != NULL)
		{
			if (FLAG_UNSET(seenf->lhisto->state, F_ESTAB))
			{
				struct inter_histo *tmparr;

				// Remove and release the length histo
				tmphisto = seenf->lhisto;
				seenf->lhisto = seenf->lhisto->next;
				free(tmphisto);

				// Remove and release the associated interarrival histo
				tmparr = seenf->thisto;
				seenf->thisto = seenf->thisto->next;
				free(tmparr);
			}
		}

		// Allocate space for the new histo, and initialize it

		if ((tmphisto = (struct len_histo *)malloc(sizeof(struct len_histo)))
			== NULL)
		{
			fprintf(gs->outf, "update_plen_histo: malloc failure, exiting!\n");
			printf("update_plen_histo: malloc failure, exiting!\n");
			exit(-1);
		}

		memset((char *)tmphisto, 0, sizeof(struct len_histo));
		
		tmphisto->histo_num = seenf->histo_num;
		tmphisto->starttime = newtime;
		tmphisto->endtime = newtime;
		tmphisto->smallestelement = MAX_LENGTH;
		tmphisto->state = flag;	// Initial state depends on why we are starting
								// a new histogram.

		tmphisto->next = seenf->lhisto;
		seenf->lhisto = tmphisto;
		seenf->num_pkts_current = 0;
	}

	// tmphisto now points to the histogram we need to update

	tmphisto->endtime = newtime;
	tmphisto->elements++;
	tmphisto->totallen += prec->len;

	// Calculate index into histogram, each bucket covers
	// gs->length_histo_width bytes

	index = prec->len/gs->length_histo_width;

	// If length is within the histogram's range, count it
	// otherwise count an 'error'

	if (index < MAX_LENGTH)
	{
		tmphisto->counts[index]++;
		if (index > tmphisto->largestelement)
			tmphisto->largestelement = index;
		if (index < tmphisto->smallestelement)
			tmphisto->smallestelement = index;
	}
	else
		tmphisto->errored_cnts++;

	return;
}


//
// update_arrival_histo()
//
// This routine is entered either to update the current
// inter-arrival time histogram, or start a new one.
//

void update_arrival_histo(struct global_state *gs,
					   struct pkt_record *prec,
					   struct seen_flow *seenf,
					   __int64 newtime,
					   int flag)
{
	struct inter_histo *tmphisto;
	long interval;	// Only used if we're updating an arrival that
					// we know is within gs->max_flow_gap, so only
					// need 'long' rather than '__int64' here

	// If 'flag' parameter is true, start a new histogram
	// If flag == F_INIT, we return immediately after
	// creating space for the new histogram - by definition
	// we're reacting either to the first ever packet for the
	// flow, or a packet so delayed that the flow was thought
	// to be previously ceased. (The other case, flag == F_ESTAB,
	// is okay since pkt flow is continuing and we're initializing
	// a new histogram to break up the statistics gathering.)

	tmphisto = seenf->thisto;
	if (flag > 0)
	{
		// Allocate space for the new histo, and initialize it

		if ((tmphisto = (struct inter_histo *)malloc(sizeof(struct inter_histo)))
			== NULL)
		{
			fprintf(gs->outf, "update_arrival_histo: malloc failure, exiting!\n");
			printf("update_arrival_histo: malloc failure, exiting!\n");
			exit(-1);
		}

		// Zero all elements
		memset((char *)tmphisto, 0, sizeof(struct inter_histo));
		
		// Initialize non-zero elements
		tmphisto->histo_num = seenf->histo_num;
		tmphisto->starttime = newtime;
		tmphisto->endtime = newtime;
		tmphisto->smallestelement = MAX_INTERARRIVAL;

		tmphisto->next = seenf->thisto;
		seenf->thisto = tmphisto;
		seenf->num_pkts_current = 0;

		if (FLAG_SET(flag, F_INIT))
			return;		// Return immediately, there's nothing meaningful
						// to record about the interval between this packet
						// and anything prior.
	}

	tmphisto->endtime = newtime;
	tmphisto->elements++;

	// tmphisto points to the histogram we need to update,
	// calculate 'interval' as an index into counts[] corresponding
	// to the bucket for X/gs->interval_histo_width

	interval = (long)((newtime - seenf->last_pkt_usec)/gs->interval_histo_width);

	// If interval is within the histogram's range, count it
	// otherwise count an 'error'

	if (interval < MAX_INTERARRIVAL)
	{
		tmphisto->counts[interval]++;
		if (interval > tmphisto->largestelement)
			tmphisto->largestelement = interval;
		if (interval < tmphisto->smallestelement)
			tmphisto->smallestelement = interval;
	}
	else
		tmphisto->errored_cnts++;

	return;
}




//
// gts_init_srv_flows()
//
// Takes a packet record, and updates the server node
// flow to which this packet record relates.
//
// The relevant seen_flow structures are assumed to have
// been established at the time the actual seen_servers
// entry is contstructed, so we don't need to check that
// each time here.
//
// As an optimization, this routine copies the seen_server's
// toserv or frmserv seen_flow entry into the packet's
// actual seen_flow entry. Subsequent packet arrival events
// can thus directly reference the toserv or frmserv seen_flow
// entry to update the to/from server flow statistics.
//

int gts_init_srv_flows(struct global_state *gs,
						 struct seen_flow *seenf,
						 struct pkt_record *prec)
{

	struct seen_servers *seensrv;
	int scnt = 0;

	seensrv = gs->ssrv;
	while (seensrv != NULL)
	{
		scnt++;

		if (((seensrv->srvaddr).s_addr == (prec->dstaddr).s_addr) &&
			(seensrv->srvport == prec->dstport))
		{
			// Pkt is heading towards this server

			if (seensrv->toserv == NULL)
			{
				// The toserv seen_flow entry doesn't yet exist,				
				// create and initialize a new seen_flow entry

				if ((seensrv->toserv = (struct seen_flow *)malloc(sizeof(struct seen_flow)))
					== NULL)
				{
					fprintf(gs->outf, "check_srv_flows: malloc failed, exiting!\n");
					printf("check_srv_flows: malloc failed, exiting!\n");
					exit(-1);
				}

				// Zero all elements of this new seen_flow entry

				memset((char *)seensrv->toserv, 0, sizeof(struct seen_flow));

				// Establish timestamp (one addr field is left zero because
				// we know, by context, only one end of the flow in this case)

				seensrv->toserv->start_sec = prec->time_sec;
				seensrv->toserv->start_usec = prec->time_usec;
				(seensrv->toserv->dstaddr).s_addr = (prec->dstaddr).s_addr;
				seensrv->toserv->dstport = prec->dstport;

				// Create ascii versions of both ip addresses for later use
				// in this and other routines

				sprintf(seensrv->toserv->saddr_txt, "%s", inet_ntoa(seensrv->toserv->srcaddr));
				sprintf(seensrv->toserv->daddr_txt, "%s", inet_ntoa(seensrv->toserv->dstaddr));

				// Pre-assign an output filename.

				sprintf(seensrv->toserv->flow_outf_name, "%s-All2Serv-%d.txt",
					gs->outf_prefix, scnt);
			}

			gts_add_flow_record(gs, prec, seensrv->toserv);

			// Finally, make sure the real flow's seen_flow entry has a pointer
			// back to the aggregate stats seen_flow entry at seensrv->toserv

			seenf->server = seensrv->toserv;

			return(scnt | TO_SERVER);

		}

		if (((seensrv->srvaddr).s_addr == (prec->srcaddr).s_addr) &&
			(seensrv->srvport == prec->srcport))
		{
			// Pkt is coming from this server

			if (seensrv->frmserv == NULL)
			{
				// The toserv seen_flow entry doesn't yet exist,				
				// create and initialize a new seen_flow entry

				if ((seensrv->frmserv = (struct seen_flow *)malloc(sizeof(struct seen_flow)))
					== NULL)
				{
					fprintf(gs->outf, "check_srv_flows: malloc failed, exiting!\n");
					printf("check_srv_flows: malloc failed, exiting!\n");
					exit(-1);
				}

				// Zero all elements of this new seen_flow entry

				memset((char *)seensrv->frmserv, 0, sizeof(struct seen_flow));

				// Establish timestamp (one addr field is left zero because
				// we know, by context, only one end of the flow in this case)

				(seensrv->frmserv->srcaddr).s_addr = (prec->srcaddr).s_addr;
				seensrv->frmserv->srcport = prec->srcport;
				seensrv->frmserv->start_sec = prec->time_sec;
				seensrv->frmserv->start_usec = prec->time_usec;

				// Create ascii versions of both ip addresses for later use
				// in this and other routines

				sprintf(seensrv->frmserv->saddr_txt, "%s", inet_ntoa(seensrv->frmserv->srcaddr));
				sprintf(seensrv->frmserv->daddr_txt, "%s", inet_ntoa(seensrv->frmserv->dstaddr));

				// Pre-assign an output filename.

				sprintf(seensrv->frmserv->flow_outf_name, "%s-Serv2All-%d.txt",
					gs->outf_prefix, scnt);
			}

			gts_add_flow_record(gs, prec, seensrv->frmserv);

			// Finally, make sure the real flow's seen_flow entry has a pointer
			// back to the aggregate stats seen_flow entry at seensrv->frmserv

			seenf->server = seensrv->frmserv;

			return(scnt | FROM_SERVER);
		}

		seensrv = seensrv->next;
	}

	return(0);
}



//
// gts_check_idle_flows()
//
// Called with the current time in sec/usec form.
//
// Checks a single flow and if it has never reached the
// F_ESTAB state and we've been through at least two
// checkpoints, completely delete the seen_flow entry
// (on the assumption that this is probably some transient
// flow that is just going to waste memory over time).
//
// Also check if the seen_flow entry's current
// histograms are effectively idle and never reached
// the F_ESTAB state. If so, the current histos are
// deleted (to conserve memory, because even if the
// flow later re-activates, the current histos would be
// superseded anyway).
//
// On entry, gs->nxt_chkf points to the seen_flow entry
// to check. Upon exit, gs->nxt_chkf points to the next
// seen_flow entry in the list hanging off gs->sflow. The
// pointer wraps back to gs->sflow when end of list is reached.
//

void gts_check_idle_flows(struct global_state *gs,
						  long sec,
						  long usec)
{
	struct len_histo *lh;
	struct inter_histo *ih;
	struct seen_flow *tmpf;
	__int64 new_relative_time;	// NB: A 'long' isn't big enough under MSVisualC++
								// (it is only 32 bits wide, same as an 'int')


	// Don't even bother if there's only one flow
	// currently in the global list.
	
	if (gs->sflow->next == NULL)
		return;											
								
	// Check if this entire flow is basically a waste of
	// space (if it never reached F_ESTAB *and* it has been at
	// least two checkpoint periods since the flow's first
	// packet was seen). If so, remove it from the seen_flow
	// list.
	
	if (FLAG_UNSET(gs->nxt_chkf->state, F_ESTAB))
	 if ((sec - gs->nxt_chkf->start_sec) > 2*gs->chkpoint_interval)
		{
			// Delete this entire seen_flow entry. Start
			// by removing any histos (there probably shouldn't
			// be any, but just in case...)
	
			while ((lh = gs->nxt_chkf->lhisto) != NULL)
			{
				gs->nxt_chkf->lhisto = gs->nxt_chkf->lhisto->next;
				free(lh);
			}
			
			while ((ih = gs->nxt_chkf->thisto) != NULL)
			{
				gs->nxt_chkf->thisto = gs->nxt_chkf->thisto->next;
				free(ih);
			}
			
			// Adjust gs->nxt_chkf to either point to the next
			// seen_flow entry or wrap around to the head of
			// the list.
			
			tmpf = gs->nxt_chkf;
			gs->nxt_chkf = gs->nxt_chkf->next;
			if (gs->nxt_chkf == NULL)
				gs->nxt_chkf = gs->sflow;
			
			// If the prev pointer is NULL then we're
			// deleting the current head of the list (!?)
			// so adjust gs->sflow.
			
			if (tmpf->prev != NULL)
				tmpf->prev->next = tmpf->next;
			else
				gs->sflow = tmpf->next;
				
			if (tmpf->next != NULL)
				tmpf->next->prev = tmpf->prev;
			
			// Free the memory
			
			free(tmpf);
			
			gs->freed_f++;	// Count this event
			
			// And return (only process one seen_flow entry each
			// time this routine is called).
			
			return;			
		}
	

	// Getting to here, we didn't delete the seen_flow so
	// check if its current histos are idle. Remove them
	// if so (the topmost len and interval histos)
	
	if (gs->nxt_chkf->lhisto != NULL)
	{
		if (FLAG_UNSET(gs->nxt_chkf->lhisto->state, F_ESTAB))
		{
			// Calculate the current time relative to the time this
			// flow started.

			new_relative_time = (__int64)1000000*(sec - gs->nxt_chkf->start_sec);
			new_relative_time += (__int64)(usec - gs->nxt_chkf->start_usec);
			if ((new_relative_time - gs->nxt_chkf->last_pkt_usec) >= gs->flow_max_gap)
			{
				// Okay, this flow is not in F_ESTAB state and has now been
				// idle for longer than the allowed gs->flow_max_gap period.
				// Remove the topmost histograms within this seen_flow entry.

				lh = gs->nxt_chkf->lhisto;
				gs->nxt_chkf->lhisto = gs->nxt_chkf->lhisto->next;
				free(lh);

				ih = gs->nxt_chkf->thisto;
				gs->nxt_chkf->thisto = gs->nxt_chkf->thisto->next;
				free(ih);
			}
		}
	}
	

	// Increment pointer to the next seen_flow entry, to be
	// inspected when gts_check_idle_flows() is next called

	gs->nxt_chkf = gs->nxt_chkf->next;
	if (gs->nxt_chkf == NULL)
		gs->nxt_chkf = gs->sflow;

	return;
}

//
// gt_flow_established()
//
// Called to note that a flow has moved from "initialized"
// to "established" state. Might be a null routine, or used
// for asynchronous alerts/screenupdates/etc.
//

void gt_flow_established(struct global_state *gs,
						struct seen_flow *seenf)
{
	long tmptime1;
	
	// Absolute start time is the start time for this actual
	// seen_flow entry plus the relative time of the histo
	// that reached "established" state (the histo currently
	// pointed to by seenf->lhisto)

	tmptime1 = seenf->start_sec +
		(long)(((__int64)seenf->start_usec + seenf->lhisto->starttime)/1000000);
	
	if (FLAG_UNSET(seenf->state, F_ESTAB))
	{
		// This is the first time the flow has reached
		// "established" state
#ifdef ARFLE
		fprintf(gs->outf, "%s:%d -> %s:%d Established at %s",
			seenf->saddr_txt, seenf->srcport,
			seenf->daddr_txt, seenf->dstport,
			ctime((time_t *)(&tmptime1)));

		printf("%s:%d -> %s:%d Est. %s",
			seenf->saddr_txt, seenf->srcport,
			seenf->daddr_txt, seenf->dstport,
			ctime((time_t *)(&tmptime1)));
#endif
	}
	else
	{
#ifdef ARFLE
		// The flow is being re-established
		fprintf(gs->outf, "%s:%d -> %s:%d Re-established at %s",
			seenf->saddr_txt, seenf->srcport,
			seenf->daddr_txt, seenf->dstport,
			ctime((time_t *)(&tmptime1)));

		printf("%s:%d -> %s:%d Re-est. %s",
			seenf->saddr_txt, seenf->srcport,
			seenf->daddr_txt, seenf->dstport,
			ctime((time_t *)(&tmptime1)));
#endif
	}

	return;
}


//
// gt_flowstart_event()
//
// Called to note the start time of a flow.
// Might be a null routine, or used for asynchronous
// alerts/screenupdates/etc.
//

void gt_flowstart_event(struct global_state *gs,
						struct seen_flow *seenf,
						long t_sec,
						long t_usec)
{
	
	return;
}



//
// gt_flowstop_event()
//
// Called to note the stop time of a flow.
// Might be a null routine, or used for asynchronous
// alerts/screenupdates/etc.
//

void gt_flowstop_event(struct global_state *gs,
						struct seen_flow *seenf,
						long t_sec,
						long t_usec)
{
	
	return;
}






