//
// gt_structs.h
//
// core structure definitions
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with pkthisto in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// NOTE on timestamps:
//
// Most of the time a 'long' appears sufficient to handle
// times in seconds and microseconds. However, under MS VisualC++
// a 'long' is 32 bits, just like an 'int'. In a couple of places
// in my code that isn't enough, so I explicitly use '__int64'
// instead to hold absolute time in microseconds. (Most of the
// rest of the program uses two 'long' variables to hold
// absolute time.)
//
// HOWEVER: "__int64" is MS-specific, and I map it to "__int64_t"
// to compile under FreeBSD4.3 gcc.
//
// the 'global_state' structure localizes all running state
// information for the statistics gathering routines

struct global_state {
	char version[16];
	char tracefile[128];		// Name of the source trace file
	int in_fd;					// Descriptor for open source trace file
	char outf_prefix[128];		// Outfile+"0.txt", intial stats file
	FILE *outf;					// File descriptor for global_outfile file
	
	char snooplink[128];		// Filename for link driver for BPF packet capture
								// (e.g. "xl0" for ethernet). If null, then the first
								// interface found on the system will be used if
								// PKTSNOOP mode has been specified.

	int debug;					// Internal debug level, used by DEBUG macro
	unsigned int stats;			// Various flags

	struct seen_flow *sflow;	// Start of list of all seen flows
	struct seen_flow *nxt_chkf;	// Pointer to next seen_flow entry to be checked
								// for idle-ness [This pointer steps by one through
								// the list off sflow each time gts_check_idle_flows()
								// is called.]
	long num_flows;				// A non-decreasing counter, increments every time a new
								// entry is added to the seen_flows list off sflow. It is
								// never decremented, and provides a unique ID number for
								// the output filenames of active flows
	long freed_f;				// Number of seen_flow entries freed due to being
								// declared idle (didn't reach F_ESTAB state in two
								// checkpoint intervals)
								
	struct seen_servers *ssrv;	// List of possible server nodes

	int timescale_factor;		// Timestamp units represent N microseconds
	
	long last_time_sec;			// Most recent timestamp seen, in seconds from 1/1/1970
								// (-1 if no packet seen yet).

	int interval_histo_width;	// Number of microseconds for each bucket in the
								// inter-arrival time histograms
	int length_histo_width;		// Number of bytes for each bucket in the length
								// histograms

	int pkt_per_histo;			// Maximum number of packets per histogram
	int pkt_cnt_min;			// Minimum number of packets in a flow before
								// we consider it "established" (we only print out
								// statistics on "established" flows)
	long total_pkts;			// If non-zero, exit after collecting
								// this many packets.
	long flow_max_gap;			// Maximum number of microseconds between two
								// packets in the same flow before that flow is
								// considered to have stopped.
	int chkpoint;				// Counts the number of checkpoints done so far
	int chkpoint_flag;			// If this is set to 1 (by a signal or other method)
								// then gts_add_pkt_record() will force a checkpoint
								// next time it is called, and reset this flag.
	long chkpoint_interval;		// Checkpoint the datasets after this many seconds
	long chkpoint_lasttime;		// Last time a checkpoint occured (seconds from 1/1/1970)


};


// As packet records are read in, if a particular flow
// source or destination appears common to multiple
// flows then store a copy of that flow's endpoint (addr/port)
// in a linked list. Most likely there's only one, but need a
// list in case the tcpdump file contains multiple concurrent
// game traffic flows). The root of this list hangs of the
// global_state structure.

struct seen_servers {
	struct in_addr srvaddr;		// IP addr of server
	int srvport;				// UDP port of server
	struct seen_flow *toserv;	// Special flow, aggregate pkts TO this node
	struct seen_flow *frmserv;	// Special flow, aggregate pkts FROM this node
	struct seen_servers *next;
};


// This struct is the top level for per-flow statistics.
// A unique entry is created for each distinct flow, and
// hanging off this entry are structures/lists representing
// packet records, length histograms, and inter-arrival time
// histograms. Also, filename for flow information outputs,
// file pointer, etc.

struct seen_flow {
	struct in_addr srcaddr;		// IP addr of flow source
	int srcport;				// UDP port of flow source
	struct in_addr dstaddr;		// IP addr of flow destination
	int dstport;				// UDP port of flow destination
	char saddr_txt[32];			// Src ipaddr in w.x.y.z format
	char daddr_txt[32];			// Dst ipaddr in w.x.y.z format

	struct seen_flow *server;	// If 'server' is non-null, this flow is
								// going to or from a specified server node.
								// This field points to the seen_flow entry
								// that is collecting aggregate flow
								// statistics to/from the server node.

	int state;					// Flags holding the state of this flow

	long start_sec;				// Timestamp when flow first began, in
	long start_usec;			// seconds and usec since Jan 1, 1970

	__int64 last_pkt_usec;		// Time of most recent packet record, in
								// usec since the beginning of this flow

	long time_established;		// Total number of seconds that this flow
								// has been in F_ESTAB state (calculated from
								// each len_histo that reaches F_ESTAB state)

	int num_pkts;				// Number of packet seen so far
	int num_pkts_current;		// Number of packet seen for current histograms

	struct len_histo *lhisto;	// Pointer to list of pkt length histograms
	struct inter_histo *thisto;	// Pointer to list of inter-arrival time histograms
	long histo_num;				// This is the Nth histogram for the flow

	char flow_outf_name[128];	// Filename for output file
	FILE *flow_outf;			// File stream descriptor for flow's output file

	struct seen_flow *prev;		// Pointer to the seen_flow entry preceding this
								// (or NULL if we're head of the list)
	struct seen_flow *next;		// Pointer to seen_flow entry following this
								// (or NULL if we're end of the list)
};


// Packet length histograms are kept as a list of histograms
// covering periods of time a few minutes long. A seen_flow struct
// may have a linked list of many len_histo structs hanging off
// it. A new histogram begins when a flow STARTS (either for the
// first time, or after previously being declared as having STOPPED).

struct len_histo {
	long histo_num;			// This is the Nth histogram for the flow (used
							// by the checkpoint routines to print a numerical
							// index uniquely identifying this histogram)
	__int64 starttime;		// Time at which this histogram begins (in 
							// usec since start of flow)
	__int64 endtime;		// Time at which this histogram ends (in 
							// usec since start of flow)
	int state;				// The state of this histo's instance
							// of the flow
	int elements;			// Number of elements in this histogram
	int errored_cnts;		// Number of lengths > 800 bytes that weren't counted
	unsigned short counts[MAX_LENGTH];
							// One bucket for each packet size from 0
							// to X bytes. VERY CRUDE. Need to improve
							// effiency once we've got the program basically
							// working on small data sets
	int totallen;			// Sum of all packet lengths (used for bandwidth estimate)
	int smallestelement;	// Smallest index into counts[] seen so far
	int largestelement;		// Largest index into counts[] seen so far
	struct len_histo *next;
};


// Inter-arrival time histograms are kept as a list of histograms
// covering periods of time a few minutes long. A seen_flow struct
// may have a linked list of many inter_histo structs hanging off
// it. A new histogram begins when a flow STARTS (either for the
// first time, or after previously being declared as having STOPPED).
// Elements represent buckets gs->inter_histo_width usec wide.
//
// NB: The 'state' of the inter-arrival histogram is assumed to be
// the same as the state of the length histogram, so there's no
// explicit state field here.

struct inter_histo {
	long histo_num;			// This is the Nth histogram for the flow (used
							// by the checkpoint routines to print a numerical
							// index uniquely identifying this histogram)
	__int64 starttime;		// Time at which this histogram begins (in 
							// usec since start of flow)
	__int64 endtime;		// Time at which this histogram ends (in 
							// usec since start of flow)
	int elements;			// Number of elements in this histogram
	int errored_cnts;		// Number of intervals > 400ms that weren't counted
	unsigned short counts[MAX_INTERARRIVAL];
							// Each bucket, gs->inter_histo_width usec wide
							// The array is crude, need to improve memory efficiency
	int smallestelement;	// Smallest index into counts[] seen so far
	int largestelement;		// Largest index into counts[] seen so far
	struct inter_histo *next;
};


// This structure is used to pass a generic record to the
// statistics routines.

struct pkt_record {
	struct in_addr srcaddr;		// Src IP addr
	struct in_addr dstaddr;		// Dst IP addr
	unsigned short srcport;		// Src UDP port
	unsigned short dstport;		// Dst UDP port
	int len;					// Actual length of IP packet (not Ethernet frame)
	long time_sec;				// Timestamp when this packet captured, measured
	long time_usec;				// in sec and usec since Jan 1, 1970
};











