//
// gt_tcpdump.cpp
//
// Routines to parse tcpdump file, and call the
// statistics gather routines appropriately
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with pkthisto in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


#include "stdafx.h"
#include "gt.h"

//
// read_tcpdump_file()
//
// Begin reading the file - check it has correct tcpdump 'magic'
// number in the header, then process subsequent frame records
// until the end of the file. For each frame record, call the
// statistics routines to update per-flow structures.
//

void read_tcpdump_file(struct global_state *gs)
{
	struct tracefile_header  header;
	struct packet_header	frame;
	int	pkteof;
	struct pkt_record prec;
	int bytesread;
	unsigned char pkt[PKT_SIZE_MIN];
	long pktcnt;

	// First check for the trace file header with tcpdump's magic
	// number

	if (_read(gs->in_fd, &header, sizeof(struct tracefile_header))
		< sizeof(struct tracefile_header))
		{
			fprintf(gs->outf,
				"read_tcpdump_file(): Error reading trace file header: %s\n",
				strerror(errno));
			exit(-1);
		}

	if (header.magic != TCPDUMP_MAGIC)
		{
			fprintf(gs->outf,
				"read_tcpdump_file(): Invalid magic number 0x08%x\n",
				header.magic);
			exit(-1);
		}

	if (header.linktype != L1_ETHERNET)
		{
			fprintf(gs->outf,
				"read_tcpdump_file(): Error, tcpdump link type %d isn't Ethernet\n",
				header.linktype);
			exit(-1);
		}

	if (FLAG_SET(gs->stats, FILE_STATS))
	{
		fprintf(gs->outf, "\nMajor version %d, Minor version %d\n",
			header.version_major, header.version_minor);

		fprintf(gs->outf, "Network type: %d   Sig Figs: %d\n",
			header.linktype, header.sigfigs);

		fprintf(gs->outf, "Snaplen: %d   This Zone: %d\n",
			header.snaplen, header.thiszone);
	}

	gs->timescale_factor = 1; // tcpdump timestamps are in microseconds

	// Now parse the following packet/frame records.

	pkteof = 1;
	pktcnt = 0;

//	while ((eof(gs->in_fd) == 0) && (pkteof != 0))
	while ((pkteof != 0))
	{

		bytesread = 0;

		// Read in the tcpdump-specific frame/packet header information

		if (_read(gs->in_fd, &frame, sizeof(struct packet_header))
			< sizeof(struct packet_header))
		{
			fprintf(gs->outf,
				"read_tcpdump_file(): EOF or Error reading packet_header: %s\n",
				strerror(errno));
			pkteof = 0;
			break;
		}

		DEBUG(gs, 1,("read_tcpdump_file(): Frame Record - processing %d bytes, actual len %d bytes\n",
			frame.capturelen, frame.actuallen));

		
		if (frame.capturelen >= PKT_SIZE_MIN)
		{
			// If the captured frame is long enough to possibly
			// contain an Ether+IP packet up to the UDP src/dst
			// port fields then we read the next 38 bytes in and
			// parse the header information.

			if (_read(gs->in_fd, &pkt, PKT_SIZE_MIN) < PKT_SIZE_MIN)
			{
				fprintf(gs->outf,
					"read_tcpdump_file(): EOF or Error reading packet_header: %s\n",
					strerror(errno));
				pkteof = 0;
				break;
			}

			// Create checkpoints at specific intervals throughout the
			// tracefile's trace period

			if ((gs->last_time_sec - gs->chkpoint_lasttime) >= gs->chkpoint_interval)
				gts_checkpoint_stats(gs, CHKP_INTERIM);

			// The IP header information is now located at index 14 onwards
			// of the array pkt[]. Ethernet header is pkt[0] to pkt[13].

			bytesread = PKT_SIZE_MIN;	// We've consumed this many
										// bytes from the input stream

			// Check to see if this is the first frame record retrieved
			// from the trace file, and if so then use its timestamp as
			// the time at which the trial started (not entirely accurate
			// in the general case, since tcpdump might have been started
			// on an idle link).
			// [Although it might appear we should do this in the stats gathering
			// routine, it is more efficient to do it here and keep the stats
			// routines free of checks that'll be needed very rarely, and
			// are tracefile-specific.]

			if (gs->last_time_sec == -1)
			{
				time_t *starttime = (time_t *)(&(frame.ts_secs));

				fprintf(gs->outf, "Trace recorded: %s\n", ctime(starttime));
				gs->last_time_sec = frame.ts_secs;
				gs->chkpoint_lasttime = frame.ts_secs;
			}

			// Fill the pkt_record

			// IP src addr is three 'words', or 12 bytes, from beginning
			// of IP header - i.e. 26 bytes from start of pkt[]

			memcpy(&(prec.srcaddr.s_addr), &(pkt[26]), 4);
			memcpy(&(prec.dstaddr.s_addr), &(pkt[30]), 4);
			prec.srcport = (pkt[34] & 0xFF)<<8  |  pkt[35];
			prec.dstport = (pkt[36] & 0xFF)<<8  |  pkt[37];

			// Copy over the length & timestamp

			prec.len = frame.actuallen - 14; // IP packet is 14 bytes shorter than Ethernet frame
			prec.time_sec = frame.ts_secs;
			prec.time_usec = frame.ts_usecs;

			// Call the statistics routine with this 'event'

			gts_add_pkt_record(gs, &prec);

		}

		// Jump pointer into tracefile forward by the number of
		// bytes we didn't consume from the current frame, to reach
		// the next frame.

		lseek(gs->in_fd, frame.capturelen - bytesread, SEEK_CUR);

	}

}

